from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing_extensions import Literal

import inspect
import types
import importlib


def import_from_string(import_path: str) -> Any:
    """Import a class from a string."""

    import_classname = import_path.split(".")[-1]
    import_module = ".".join(import_path.split(".")[:-1])

    module = importlib.import_module(import_module)
    return getattr(module, import_classname)


def get_input_args():
    frame = None
    current_frame = inspect.currentframe()

    if current_frame:
        frame = current_frame.f_back

    if not isinstance(frame, types.FrameType):
        raise AttributeError("There is no `frame` available while being required.")

    init_args = {}
    for local_args in collect_init_args(frame, []):
        init_args.update(local_args)

    return init_args


# NOTE(hadim): the below is coming from PL.
# https://github.com/PyTorchLightning/pytorch-lightning/blob/87bd54aedfcd990b935860d020ca89d192e2ba40/pytorch_lightning/utilities/parsing.py#L161


def parse_class_init_keys(cls) -> Tuple[str, Optional[str], Optional[str]]:
    """Parse key words for standard ``self``, ``*args`` and ``**kwargs``.
    Examples:
        >>> class Model():
        ...     def __init__(self, hparams, *my_args, anykw=42, **my_kwargs):
        ...         pass
        >>> parse_class_init_keys(Model)
        ('self', 'my_args', 'my_kwargs')
    """
    init_parameters = inspect.signature(cls.__init__).parameters
    # docs claims the params are always ordered
    # https://docs.python.org/3/library/inspect.html#inspect.Signature.parameters
    init_params = list(init_parameters.values())
    # self is always first
    n_self = init_params[0].name

    def _get_first_if_any(
        params: List[inspect.Parameter],
        param_type: Literal[
            inspect._ParameterKind.VAR_POSITIONAL, inspect._ParameterKind.VAR_KEYWORD
        ],
    ) -> Optional[str]:
        for p in params:
            if p.kind == param_type:
                return p.name
        return None

    n_args = _get_first_if_any(init_params, inspect.Parameter.VAR_POSITIONAL)
    n_kwargs = _get_first_if_any(init_params, inspect.Parameter.VAR_KEYWORD)

    return n_self, n_args, n_kwargs


def get_init_args(frame: types.FrameType) -> Dict[str, Any]:
    _, _, _, local_vars = inspect.getargvalues(frame)
    if "__class__" not in local_vars:
        return {}
    cls = local_vars["__class__"]
    init_parameters = inspect.signature(cls.__init__).parameters
    self_var, args_var, kwargs_var = parse_class_init_keys(cls)
    filtered_vars = [n for n in (self_var, args_var, kwargs_var) if n]
    exclude_argnames = (*filtered_vars, "__class__", "frame", "frame_args")
    # only collect variables that appear in the signature
    local_args = {k: local_vars[k] for k in init_parameters.keys()}
    # kwargs_var might be None => raised an error by mypy
    if kwargs_var:
        local_args.update(local_args.get(kwargs_var, {}))
    local_args = {k: v for k, v in local_args.items() if k not in exclude_argnames}
    return local_args


def collect_init_args(
    frame: types.FrameType, path_args: List[Dict[str, Any]], inside: bool = False
) -> List[Dict[str, Any]]:
    """Recursively collects the arguments passed to the child constructors in the inheritance tree.
    Args:
        frame: the current stack frame
        path_args: a list of dictionaries containing the constructor args in all parent classes
        inside: track if we are inside inheritance path, avoid terminating too soon
    Return:
          A list of dictionaries where each dictionary contains the arguments passed to the
          constructor at that level. The last entry corresponds to the constructor call of the
          most specific class in the hierarchy.
    """
    _, _, _, local_vars = inspect.getargvalues(frame)
    # frame.f_back must be of a type types.FrameType for get_init_args/collect_init_args due to mypy
    if not isinstance(frame.f_back, types.FrameType):
        return path_args

    if "__class__" in local_vars:
        local_args = get_init_args(frame)
        # recursive update
        path_args.append(local_args)
        return collect_init_args(frame.f_back, path_args, inside=True)
    if not inside:
        return collect_init_args(frame.f_back, path_args, inside)
    return path_args



from typing import Union, List, Optional

import datamol as dm
import numpy as np

from loguru import logger
from numpy.linalg import norm
from scipy.special import cbrt
from rdkit.Chem import rdPartialCharges, rdForceFieldHelpers
from rdkit.Chem import rdMolDescriptors
from molfeat.calc.base import SerializableCalculator
from molfeat.utils.commons import requires_conformer


class USRDescriptors(SerializableCalculator):
    """Descriptors for the shape of a molecule.

    !!! note:
        The following shape descriptors are offered:
            * USR: UltraFast Shape Recognition
            * USRCAT: Ultrafast Shape Recognition with CREDO Atom Types
    """

    def __init__(self, method: str = "USR", replace_nan: bool = False, **kwargs):
        """Constructor for ShapeDescriptors

        Args:
            method: Shape descriptor method to use. One of 'USR', 'USRCAT'. Default to 'USR'
            replace_nan: Whether to replace nan or infinite values. Defaults to False.
        """
        self.method = method.upper()
        if self.method not in ["USR", "USRCAT"]:
            raise ValueError(f"Shape descriptor {self.method} is not supported")
        self.replace_nan = replace_nan
        self._columns = None

    def __getstate__(self):
        state = {}
        state["method"] = self.method
        state["replace_nan"] = self.replace_nan
        state["_columns"] = self._columns
        return state

    @property
    def columns(self):
        """
        Get the name of all the descriptors of this calculator
        """
        if self._columns is None:
            if self.method == "USR":
                self._columns = [f"usr-{i}" for i in range(1, 13)]
            elif self.method == "USRCAT":
                self._columns = [f"usr-{i}" for i in range(1, 61)]
        return self._columns

    def __len__(self):
        """Compute descriptors length"""
        return len(self.columns)

    @requires_conformer
    def __call__(self, mol: Union[dm.Mol, str], conformer_id: Optional[int] = -1):
        r"""
        Get rdkit 3D descriptors for a molecule

        Args:
            mol: the molecule of interest
            conformer_id (int, optional): Optional conformer id. Defaults to -1.

        Returns:
            shape_descriptors (np.ndarray): list of computed mordred molecular descriptors
        """
        if self.method == "USR":
            shape_descr = rdMolDescriptors.GetUSR(mol, confId=conformer_id)
        elif self.method == "USRCAT":
            shape_descr = rdMolDescriptors.GetUSRCAT(mol, confId=conformer_id)
        if self.replace_nan:
            shape_descr = np.nan_to_num(shape_descr, self.replace_nan)
        return np.asarray(shape_descr)


class ElectroShapeDescriptors(SerializableCalculator):
    """Compute Electroshape descriptors as described by

    Armstrong et al. ElectroShape: fast molecular similarity calculations incorporating shape, chirality and electrostatics.
    J Comput Aided Mol Des 24, 789-801 (2010). http://dx.doi.org/doi:10.1007/s10822-010-9374-0
    """

    SUPPORTED_CHARGE_MODELS = ["gasteiger", "tripos", "mmff94", "formal"]

    def __init__(
        self,
        charge_model: str = "gasteiger",
        replace_nan: bool = False,
        electron_scaling: float = 25.0,
        **kwargs,
    ):
        """Constructor for ElectroShape descriptor

        Args:
            charge_model: charge model to use. One of ('gasteiger', 'tripos', 'mmff94', 'formal'). Defaults to "gasteiger".
                Note that formal charges will be computed on the fly if not provided in the input molecules.
                The `tripos` charge models comes from TRIPOS force field and are often parsed from mol2 files.
            replace_nan: whether to replace NaN values. Defaults False
            electron_scaling: scaling factor to convert electron charges to Angstroms. Defaults to 25.0.
        """

        self.charge_model = charge_model
        self.replace_nan = replace_nan
        self.electron_scaling = electron_scaling
        self._columns = None

    @property
    def columns(self):
        """
        Get the name of all the descriptors of this calculator
        """
        if self._columns is None:
            self._columns = []
            for i in range(1, 6):
                self._columns.extend([f"dist-{i}-mean", f"dist-{i}-std", f"dist-{i}-crb"])

        return self._columns

    def __getstate__(self):
        state = {}
        state["charge_model"] = self.charge_model
        state["replace_nan"] = self.replace_nan
        state["electron_scaling"] = self.electron_scaling
        state["_columns"] = self._columns
        return state

    def __len__(self):
        """Return the length of the calculator"""
        return len(self.columns)

    @staticmethod
    def compute_charge(mol: Union[dm.Mol, str], charge_model: str = None):
        """
        Get the molecular charge of the molecule.

        Args:
            charge_model: charge model to use. One of ('gasteiger', 'tripos', 'mmff94', 'formal'). Defaults to "gasteiger".
        """

        if charge_model not in ElectroShapeDescriptors.SUPPORTED_CHARGE_MODELS:
            raise ValueError(
                f"Unknown charge model {charge_model}. You should provide one of {ElectroShapeDescriptors.SUPPORTED_CHARGE_MODELS}"
            )
        mol = dm.to_mol(mol)
        atom_charge = []
        atom_list = list(mol.GetAtoms())

        # force compute the partial charges if not provided
        if charge_model == "gasteiger" and not atom_list[0].HasProp("_GasteigerCharge"):
            rdPartialCharges.ComputeGasteigerCharges(mol)
        elif charge_model == "mmff94" and not atom_list[0].HasProp("_MMFF94Charge"):
            ff_infos = rdForceFieldHelpers.MMFFGetMoleculeProperties(mol)
            for i, atom in enumerate(atom_list):
                atom.SetDoubleProp("_MMFF94Charge", ff_infos.GetMMFFPartialCharge(i))

        for atom in mol.GetAtoms():
            if charge_model == "formal":
                atom_charge.append(atom.GetFormalCharge())
            elif charge_model == "gasteiger":
                atom_charge.append(atom.GetDoubleProp("_GasteigerCharge"))
            elif charge_model == "mmff94":
                atom_charge.append(atom.GetDoubleProp("_MMFF94Charge"))
            elif charge_model == "tripos":
                atom_charge.append(atom.GetDoubleProp("_TriposPartialCharge"))
        return np.asarray(atom_charge)

    @requires_conformer
    def __call__(self, mol: Union[dm.Mol, str], conformer_id: Optional[int] = -1):
        r"""
        Get rdkit 3D descriptors for a molecule

        Args:
            mol: the molecule of interest
            conformer_id (int, optional): Optional conformer id. Defaults to -1.

        Returns:
            shape_descriptor (np.ndarray): computed shape descriptor
        """

        mol = dm.to_mol(mol)
        coords = mol.GetConformer(conformer_id).GetPositions()
        charge = self.compute_charge(mol, self.charge_model)
        if self.replace_nan:
            charge = np.nan_to_num(charge)

        desc_4d = np.column_stack((coords, charge * self.electron_scaling))

        c1 = desc_4d.mean(axis=0)
        distances_c1 = norm(desc_4d - c1, axis=1)

        c2 = desc_4d[distances_c1.argmax()]  # atom position furthest from c1
        distances_c2 = norm(desc_4d - c2, axis=1)

        c3 = desc_4d[distances_c2.argmax()]  # atom position furthest from c2
        distances_c3 = norm(desc_4d - c3, axis=1)

        vector_a = c2 - c1
        vector_b = c3 - c1
        vector_as = vector_a[:3]  # spatial parts of these vectors
        vector_bs = vector_b[:3]  # spatial parts of these vectors
        cross_ab = np.cross(vector_as, vector_bs)
        vector_c = (norm(vector_a) / (2 * norm(cross_ab))) * cross_ab
        vector_c1s = c1[:3]

        max_charge = np.array(np.amax(charge) * self.electron_scaling)
        min_charge = np.array(np.amin(charge) * self.electron_scaling)

        c4 = np.append(vector_c1s + vector_c, max_charge)
        c5 = np.append(vector_c1s + vector_c, min_charge)

        distances_c4 = norm(desc_4d - c4, axis=1)
        distances_c5 = norm(desc_4d - c5, axis=1)

        distances_list = [
            distances_c1,
            distances_c2,
            distances_c3,
            distances_c4,
            distances_c5,
        ]

        shape_descriptor = np.zeros(15)

        i = 0
        for distances in distances_list:
            mean = np.mean(distances)
            shape_descriptor[0 + i] = mean
            shape_descriptor[1 + i] = np.std(distances)
            shape_descriptor[2 + i] = cbrt(np.sum(((distances - mean) ** 3) / distances.size))
            i += 3
        if self.replace_nan:
            return np.nan_to_num(shape_descriptor)
        return shape_descriptor


def usrdistance(
    shape_1,
    shape_2,
    weights: Optional[List[float]] = None,
):
    """Computes similarity between molecules

    Args:
        shape_1: USR shape descriptor of first molecule
        shape_2: USR shape descriptor
        weights: List of scaling factor to use for

    Returns:
        dist: Distance [0-1] between shapes of molecules, 0 indicates identical molecules
    """

    # case for usr shape descriptors
    if weights is None:
        weights = []
    if (
        (shape_1.shape[-1] == shape_2.shape[-1] == 12)
        or (shape_1.shape[-1] == shape_2.shape[-1] == 60)
        or (shape_1.shape[-1] == shape_2.shape[-1] == 15)
    ):
        dist = rdMolDescriptors.GetUSRScore(shape_1, shape_2, weights=weights)
        return dist

    raise Exception(
        "Given vectors are not valid USR shape descriptors "
        "or come from different methods. Correct vector lengths"
        "are: 12 for USR, 60 for USRCAT, 15 for Electroshape"
    )



from typing import List
from typing import Callable
from typing import Optional

import re
import numpy as np
import torch

from molfeat.utils import datatype
from molfeat.utils.pooler import Pooling
from molfeat.trans.base import MoleculeTransformer
from molfeat.utils.commons import _parse_to_evaluable_str


class ESMProteinFingerprint(MoleculeTransformer):
    """
    ESM (Evolutionary Scale Modeling) protein representation embedding.
    ESM is a transformer protein language model introduced by Facebook FAIR in Rives et al., 2019:
    'Biological structure and function emerge from scaling unsupervised learning to 250 million protein sequences'
    """

    def __init__(
        self,
        featurizer: str = "esm1b_t33_650M_UR50S",
        loader_repo_or_dir: str = "facebookresearch/esm:main",
        device: Optional[str] = None,
        layers: List[int] = None,
        pooling: str = "mean",
        dtype: Callable = None,
        contact: bool = False,
        **kwargs,
    ):
        """Constructor for ESM protein representation

        Args:
            featurizer: Name of the ESM model to use. Defaults to "esm1b_t33_650M_UR50S".
            loader_repo_or_dir: Path to local dir containing the model or to a github repo. Default to "facebookresearch/esm:main
            device: Torch device to move the model to. Defaults to None.
            layers: Layers to use to extract information. Defaults to None, which is the last layers.
            pooling: Pooling method to use for sequence embedding. Defaults to "mean".
                If you set pooling to None, token representation will be returned (excluding BOS)
            dtype: Representation output datatype. Defaults to None.
            contact: Whether to return the predictied attention contact instead of the representation. Defaults to False.
        """
        self._model_name = featurizer
        self.device = device
        self.dtype = dtype
        self.featurizer = None
        self.alphabet = None
        self.batch_converter = None
        self._fitted = True
        self.cols_to_keep = None
        self.repr_layers = layers
        self.repo_or_dir = loader_repo_or_dir
        self.contact = contact
        max_layer_pattern = re.compile(".*_t([0-9]+)_.*")
        self._max_layers = int(max_layer_pattern.match(featurizer).group(1))
        if layers is None:
            self.repr_layers = [self._max_layers]
        if any(l > self._max_layers for l in self.repr_layers):
            raise ValueError(
                "You are requesting more layers than available for this pretrained model"
            )
        self._representation = "seq"
        self.pooling = Pooling(dim=0, name=pooling)
        if pooling is None:
            self._representation = "token"
        self._feat_length = None
        self._load_model()

    def _load_model(self):
        """Load model internally"""
        self.featurizer, self.alphabet = torch.hub.load(self.repo_or_dir, self._model_name)  # type: ignore
        self.batch_converter = self.alphabet.get_batch_converter()
        if self.device is not None:
            self.featurizer = self.featurizer.to(self.device)
        self.featurizer.eval()

    def __len__(self):
        """Get featurizer length"""
        if self._feat_length is None and not self.contact:
            embds = self._transform("MMMM")
            self._feat_length = embds.shape[-1]
        return self._feat_length

    @property
    def n_layers(self):
        """Number of layers used in the current embeddings"""
        return len(self.repr_layers)

    @torch.no_grad()
    def _embed(self, prot_seqs: List[str], prot_names: Optional[List[str]] = None, **kwargs):
        r"""
        Compute features for a single molecule.
        This method would potentially need to be reimplemented by child classes

        Args:
           prot_seqs: protein sequences as a sequence of amino acids
           prot_names: protein names

        Returns
            feat: list of N_SEQ representation, each of size (SEQ_LEN, FEAT_DIM * N_LAYERS) for token embeddings
                and (FEAT_DIM * N_LAYERS) for sequence embeddings. Note that SEQ_LEN will include the stop token.

        """
        if isinstance(prot_seqs, str):
            prot_seqs = [prot_seqs]
        if prot_names is None:
            prot_names = ["protein_{i}" for i in range(len(prot_seqs))]
        if isinstance(prot_names, str):
            prot_names = [prot_names]
        if len(prot_seqs) != len(prot_names):
            raise ValueError("Must provide the same number of protein sequence and label")
        data = list(zip(prot_names, prot_seqs))
        *_, batch_tokens = self.batch_converter(data)
        if self.device is not None:
            batch_tokens = batch_tokens.to(self.device)

        results = self.featurizer(
            batch_tokens, repr_layers=self.repr_layers, return_contacts=self.contact
        )
        embeddings = []
        if self.contact:
            for _, (seq, att_concats) in enumerate(zip(prot_seqs, results["contacts"])):
                embeddings.append(att_concats[: len(seq), : len(seq)])
        else:
            representation = torch.stack(
                [results["representations"][x] for x in self.repr_layers], dim=-1
            )
            if self._representation.startswith("seq"):
                for seq, token_rep in zip(prot_seqs, representation):
                    embeddings.append(
                        self.pooling(token_rep[1 : len(seq) + 1]).view(1, -1).squeeze(0)
                    )
            else:
                embeddings = list(
                    representation.view(representation.shape[0], representation.shape[1], -1)
                )
        return embeddings

    def __repr__(self):
        return "{}(model={}, pooling={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self._model_name),
            _parse_to_evaluable_str(self.pooling.name),
            _parse_to_evaluable_str(self.dtype),
        )

    def _transform(self, protein_seq: str, protein_name: str = None):
        """
        Transform a protein sequence into a feature vector.

        Args:
            protein: protein sequence as amino acid sequences
            protein_name: protein name

        Returns:
            Embedding of size (SEQ_LEN, FEAT_DIM, N_LAYERS) for token embeddings
                and (FEAT_DIM * N_LAYERS) for sequence embeddings
        """
        return self._embed(protein_seq, protein_name)[0]

    def transform(self, seqs: List[str], names: Optional[List[str]] = None, **kwargs):
        """
        Transform a list of protein sequence into a feature vector.

        Args:
            seqs: list of protein sequence as amino acids
            names: protein names

        Returns:
            Embedding of size (N_SEQS, SEQ_LEN, FEAT_DIM * N_LAYERS) for token embeddings
                and (N_SEQS, FEAT_DIM * N_LAYERS) for sequence embeddings. Use
        """
        if (
            names is None
            and isinstance(seqs, list)
            and isinstance(seqs[0], list)
            and len(seqs[0]) == 2
        ):
            names, seqs = zip(*seqs)
            seqs = list(seqs)
            names = list(names)
        return self._embed(seqs, names)

    def __call__(
        self,
        seqs: List[str],
        names: Optional[List[str]] = None,
        ignore_errors: bool = False,
        enforce_dtype: bool = True,
        **kwargs,
    ):
        r"""
        Compute molecular representation of a protein sequence.
        If ignore_error is True, a list of features and valid ids are returned.

        Args:
            seqs: list of protein sequence as amino acids
            names: protein names
            enforce_dtype: whether to enforce the instance dtype in the generated fingerprint
            ignore_errors: Whether to ignore errors during featurization or raise an error.
            kwargs: Named parameters for the transform method

        Returns:
            feats: list of valid embeddings
            ids: all valid positions that did not failed during featurization.
                Only returned when ignore_errors is True.

        """
        features = self.transform(seqs, names, ignore_errors=ignore_errors, **kwargs)
        ids = np.arange(len(features))
        if ignore_errors:
            features, ids = self._filter_none(features)
        if self.dtype is not None and enforce_dtype:
            if self.contact or not self._representation.startswith("seq"):
                features = [
                    datatype.cast(feat, dtype=self.dtype, columns=self.columns) for feat in features
                ]
            else:
                features = datatype.cast(features, dtype=self.dtype, columns=self.columns)
        if not ignore_errors:
            return features
        return features, ids



import os
import shutil
import unittest as ut
import datamol as dm
import pandas as pd
import torch
import numpy as np
import shelve
import tempfile
import h5py
import joblib
from molfeat.utils import datatype
from molfeat.utils import commons
from molfeat.utils.cache import CacheList, DataCache
from molfeat.utils.cache import FileCache
from molfeat.trans.fp import FPVecTransformer


def utils_fn2(x):
    return x**2


class TestUtils(ut.TestCase):
    def test_pickle_fn(self):
        fn1 = lambda x: x + 1
        fn3 = np.log10
        fn2 = utils_fn2
        hex1 = commons.fn_to_hex(fn1)
        hex2 = commons.fn_to_hex(fn2)
        hex3 = commons.fn_to_hex(np.log10)  # we are testing without var assign
        inp = 10
        r_fn1 = commons.hex_to_fn(hex1)
        r_fn2 = commons.hex_to_fn(hex2)
        r_fn3 = commons.hex_to_fn(hex3)
        self.assertListEqual([fn1(inp), fn2(inp), fn3(inp)], [r_fn1(inp), r_fn2(inp), r_fn3(inp)])
        with self.assertRaises(AttributeError):
            # we cannot pickle local function
            # this is impossible by design
            def fn2(x):
                return x**2

            hex2 = commons.fn_to_hex(fn2)

    def test_dtype(self):
        self.assertTrue(datatype.is_dtype_tensor(torch.int))
        self.assertFalse(datatype.is_dtype_tensor(int))
        self.assertTrue(datatype.is_dtype_numpy(np.float32))

    def test_null(self):
        self.assertTrue(datatype.is_null(None))
        self.assertTrue(datatype.is_null([float("nan"), np.NAN]))
        self.assertFalse(datatype.is_null([float("nan"), 1.0]))

    def test_cast(self):
        arr1 = np.random.randn(5, 1)
        dict1 = dict(test=np.array([2, 3, 1]))
        torch_arr1 = datatype.cast(arr1, torch.float)
        self.assertTrue(np.allclose(arr1, torch_arr1.cpu().numpy()))
        arr2 = None
        self.assertIsNone(datatype.cast(arr2, int))
        self.assertListEqual(list(datatype.cast(dict1, list)["test"]), list(dict1["test"]))

    def test_one_hot(self):
        val1 = 1
        val2 = 10
        allowable_set = range(5)
        enc1 = commons.one_hot_encoding(val1, allowable_set)
        enc2 = commons.one_hot_encoding(val2, allowable_set, encode_unknown=True)
        self.assertListEqual(list(enc1), [0, 1, 0, 0, 0])
        self.assertListEqual(list(enc2), [0, 0, 0, 0, 0, 1])


class TestCache(ut.TestCase):
    def test_datacache(self):
        mol_data = dm.data.freesolv()
        # in memory cache
        smiles_list = mol_data["smiles"].values
        cache = DataCache(name="test", delete_on_exit=True)
        featurizer = FPVecTransformer(kind="rdkit", length=10)
        expected_output = datatype.to_numpy(featurizer.transform(smiles_list))
        computed_data = datatype.to_numpy(cache(smiles_list, featurizer))
        np.testing.assert_array_equal(expected_output, computed_data)

        # refetch data from cache
        refetched_data = cache.fetch(smiles_list)
        refetched_data = datatype.to_numpy(refetched_data)
        np.testing.assert_array_equal(expected_output, refetched_data)

        # test cache on local storage with shelve
        disk_cache = DataCache(name="test2", cache_file=True, delete_on_exit=True)
        self.assertIsInstance(disk_cache.cache, shelve.Shelf)
        computed_data2 = datatype.to_numpy(disk_cache(smiles_list, featurizer))
        np.testing.assert_array_equal(expected_output, computed_data2)

        # check data retrieval
        first_smiles_val = expected_output[0]
        first_smiles = smiles_list[0]
        first_mol = dm.to_mol(first_smiles)
        # check data
        self.assertTrue(first_smiles in disk_cache)
        self.assertFalse("FAKE" in disk_cache)

        np.testing.assert_array_equal(disk_cache[first_smiles], first_smiles_val)
        np.testing.assert_array_equal(disk_cache[first_mol], first_smiles_val)

        # test saving and reloading cache
        save_file = tempfile.NamedTemporaryFile(delete=False, suffix=".pkl")
        save_file = save_file.name
        disk_cache.save_to_file(save_file)
        new_cache = DataCache.load_from_file(save_file)
        self.assertTrue(first_smiles in new_cache)
        np.testing.assert_array_equal(new_cache[first_smiles], first_smiles_val)
        try:
            os.unlink(save_file)
        except:
            pass

    def test_filecache(self):
        mol_data = dm.data.freesolv().iloc[:100]
        # in memory cache
        smiles_list = mol_data["smiles"].values
        hash_fn = dm.unique_id
        featurizer = FPVecTransformer(kind="rdkit", length=10)
        vals = datatype.to_numpy(featurizer.transform(smiles_list))
        mol_ids = [hash_fn(x) for x in smiles_list]

        first_smiles_val = vals[0]
        first_smiles = smiles_list[0]
        first_mol = dm.to_mol(first_smiles)

        precomputed_data = dict(zip(mol_ids, vals))
        # TEST USING A DATAFRAME
        df = pd.DataFrame(precomputed_data.items(), columns=["keys", "feats"])
        df["values"] = df["feats"].apply(commons.pack_bits)
        with tempfile.NamedTemporaryFile(delete=True, suffix="csv") as temp_file:
            df.to_csv(temp_file, index=False)
            cache = FileCache(temp_file.name, mol_hasher=hash_fn, file_type="csv")
            # check data
            self.assertTrue(first_smiles in cache)
            self.assertFalse("FAKE" in cache)
            np.testing.assert_array_equal(cache[first_smiles], first_smiles_val)

        with tempfile.NamedTemporaryFile(delete=True, suffix=".pkl") as temp_file:
            joblib.dump(precomputed_data, temp_file.name)
            cache_pkl = FileCache(temp_file.name, mol_hasher=hash_fn, file_type="pickle")
            # check data
            self.assertTrue(first_mol in cache_pkl)
            self.assertFalse("FAKE" in cache_pkl)
            np.testing.assert_array_equal(cache_pkl[first_smiles], first_smiles_val)

        with tempfile.NamedTemporaryFile(delete=True, suffix=".parquet") as temp_file:
            df_parquet = pd.DataFrame(precomputed_data.items(), columns=["keys", "feats"])
            df_parquet["values"] = df_parquet["feats"]
            df_parquet.to_parquet(temp_file.name)
            cache = FileCache(temp_file.name, mol_hasher=hash_fn, file_type="parquet")
            # check data
            self.assertTrue(first_mol in cache)
            self.assertFalse("FAKE" in cache)
            np.testing.assert_array_equal(cache[first_smiles], first_smiles_val)

        with tempfile.NamedTemporaryFile(delete=True, suffix=".h5") as temp_file:
            with h5py.File(temp_file, "w") as f:
                for k, v in precomputed_data.items():
                    f.create_dataset(k, data=v)

            cache = FileCache(temp_file.name, mol_hasher=hash_fn, file_type="hdf5")
            # check data
            self.assertTrue(first_smiles in cache)
            self.assertFalse("FAKE" in cache)
            self.assertEqual(len(cache), len(smiles_list))
            np.testing.assert_array_equal(cache[first_smiles], first_smiles_val)

            # refetch data from cache
            refetched_data = cache.fetch(smiles_list)
            refetched_data = datatype.to_numpy(refetched_data)
            np.testing.assert_array_equal(vals, refetched_data)

        # check cache updating and reloading reloading of cache
        with tempfile.NamedTemporaryFile(delete=True, suffix=".pkl") as temp_file:
            smiles_list2 = dm.data.freesolv()["smiles"].iloc[50:150].values
            _ = cache(smiles_list2, featurizer)
            self.assertEqual(len(cache), 150)
            cache.save_to_file(temp_file.name, file_type="pickle")
            parquet_out = temp_file.name + ".parquet"
            csv_out = temp_file.name + ".csv"
            cache.save_to_file(parquet_out, file_type="parquet")
            cache.save_to_file(csv_out, file_type="csv")
            reloaded_cache = FileCache.load_from_file(temp_file.name, file_type="pickle")
            reloaded_cache_parquet = FileCache.load_from_file(parquet_out, file_type="parquet")
            reloaded_cache_csv = FileCache.load_from_file(csv_out, file_type="csv")
            self.assertTrue(first_smiles in reloaded_cache, msg=reloaded_cache)
            self.assertFalse("FAKE" in reloaded_cache)
            np.testing.assert_array_equal(reloaded_cache[first_smiles], first_smiles_val)
            np.testing.assert_array_equal(reloaded_cache_parquet[first_smiles], first_smiles_val)
            np.testing.assert_array_equal(reloaded_cache_csv[first_smiles], first_smiles_val)
            for path in [parquet_out, csv_out]:
                try:
                    os.unlink(path)
                except:
                    shutil.rmtree(path)

    def test_cache_list(self):
        # Test multiple cache simultaneously
        mol_data = dm.data.freesolv().iloc[:200]
        smiles_list = mol_data["smiles"].values
        featurizer = FPVecTransformer(kind="rdkit", length=10)
        vals = datatype.to_numpy(featurizer.transform(smiles_list))

        with tempfile.NamedTemporaryFile(delete=True, suffix="csv") as temp_file:
            cache1 = FileCache(temp_file.name, file_type="csv")
            cache1(smiles_list[:50], featurizer)

            cache2 = DataCache(name="test2", cache_file=True, delete_on_exit=True)
            cache2(smiles_list[50:100], featurizer)

            cache3 = DataCache(name="test2", cache_file=False, delete_on_exit=True)
            cache3(smiles_list[100:150], featurizer)

            cache_merge = CacheList(cache1, cache2)
            np.testing.assert_array_equal(cache_merge.fetch(smiles_list)[:100], vals[:100])

            cache_merge.update(cache3)
            np.testing.assert_array_equal(cache_merge.fetch(smiles_list)[:150], vals[:150])

            cache_merge.clear()
            self.assertTrue(len(cache1) == 0)
            self.assertTrue(len(cache2) == 0)
            # this one should not change
            self.assertTrue(len(cache3) != 0)

    def test_align_conformers(self):
        # Get some molecules
        smiles_list = [
            "Nc1cnn(-c2ccccc2)c(=O)c1Cl",
            "Cc1ccn(-c2ccccc2)c(=O)c1F",
            "Cc1cnn(-c2ccccc2)c(=O)c1Cl",
            "Cc1cnn(-c2ccccc2)c(=O)c1",
        ]
        mols = [dm.to_mol(smiles) for smiles in smiles_list]

        # Generate conformers
        mols = [dm.conformers.generate(mol) for mol in mols]

        # Save average coordinates (just for the assert below)
        dummy_pos_averages = [dm.conformers.get_coords(mol).mean() for mol in mols]

        # Align conformers
        mols, scores = commons.align_conformers(mols)

        # Get the average coordinates after aligning
        dummy_pos_averages_aligned = [dm.conformers.get_coords(mol).mean() for mol in mols]

        # Reference mol should have the same coordinate
        assert np.allclose(dummy_pos_averages[:1], dummy_pos_averages_aligned[:1])

        # Other molecules should have different coordinates
        for p1, p2 in zip(dummy_pos_averages[1:], dummy_pos_averages_aligned[1:]):
            assert p1 != p2



import unittest as ut
import time
import numpy as np
import datamol as dm
import torch
import pytest
import tempfile
import joblib
from molfeat.trans.pretrained import GraphormerTransformer
from molfeat.trans.pretrained import PretrainedDGLTransformer
from molfeat.trans.pretrained import PretrainedHFTransformer
from molfeat.utils import requires
from molfeat.utils.cache import DataCache


@pytest.mark.xfail(
    not requires.check("graphormer_pretrained"), reason="3rd party module graphormer is missing"
)
class TestGraphormerTransformer(ut.TestCase):
    r"""Test cases for FingerprintsTransformer"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
    ]

    def test_graphormer_pickling(self):
        transf = GraphormerTransformer(dtype=np.float32, pooling="sum")
        with tempfile.NamedTemporaryFile(delete=True) as pickled_file:
            joblib.dump(transf, pickled_file.name)
            trans2 = joblib.load(pickled_file.name)
        ori_feat = transf(dm.freesolv().smiles.values[:10])
        reloaded_feat = trans2(dm.freesolv().smiles.values[:10])
        np.testing.assert_array_equal(ori_feat, reloaded_feat)

    def test_graphormer(self):
        transf = GraphormerTransformer(dtype=np.float32, pooling="sum")
        trans2 = GraphormerTransformer(dtype=np.float32, pooling="mean")
        fps = transf(self.smiles, enforce_dtype=True)
        fps2 = transf(self.smiles, enforce_dtype=True)
        fps3 = trans2(self.smiles, enforce_dtype=True)
        self.assertEqual(len(fps), 3)
        self.assertEqual(len(transf), fps[0].shape[-1])
        np.testing.assert_array_equal(fps, fps2)
        node_per_mol = np.rint(fps / fps3)
        node_per_mol = np.unique(node_per_mol)
        self.assertLessEqual(len(node_per_mol), 2)

    def test_graphormer_cache(self):
        transf = GraphormerTransformer(
            dtype=np.float32, pooling="mean", max_length=25, precompute_cache=True
        )
        t0 = time.time()
        time_buffer = 1
        fps = transf.transform(self.smiles)
        ori_run = time.time() - t0
        fps2 = transf.transform(self.smiles)
        cached_run = time.time() - t0 - ori_run
        np.testing.assert_array_equal(fps, fps2)
        # add buffers
        self.assertLessEqual(cached_run, ori_run + time_buffer)


class TestDGLTransformer(ut.TestCase):
    r"""Test cases for FingerprintsTransformer"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
    ]

    def test_dgl_pickling(self):
        transf = PretrainedDGLTransformer(dtype=np.float32, pooling="sum")
        with tempfile.NamedTemporaryFile(delete=True) as pickled_file:
            joblib.dump(transf, pickled_file.name)
            trans2 = joblib.load(pickled_file.name)
        ori_feat = transf(dm.freesolv().smiles.values[:10])
        reloaded_feat = trans2(dm.freesolv().smiles.values[:10])
        np.testing.assert_array_equal(ori_feat, reloaded_feat)

    def test_dgl_pretrained(self):
        transf = PretrainedDGLTransformer(dtype=np.float32, pooling="sum")
        fps = transf(self.smiles, enforce_dtype=True)
        fps2 = transf(self.smiles, enforce_dtype=True)
        self.assertEqual(len(fps), 3)
        self.assertEqual(len(transf), fps[0].shape[-1])
        np.testing.assert_array_equal(fps, fps2)

    @pytest.mark.xfail(reason="Cache might not be faster")
    def test_dgl_pretrained_cache(self):
        transf = PretrainedDGLTransformer(
            dtype=np.float32, pooling="mean", max_length=25, precompute_cache=True
        )
        t0 = time.time()
        time_buffer = 1
        fps = transf.transform(self.smiles)
        ori_run = time.time() - t0
        fps2 = transf.transform(self.smiles)
        cached_run = time.time() - t0 - ori_run
        np.testing.assert_array_equal(fps, fps2)
        # add buffers
        self.assertLessEqual(cached_run, ori_run + time_buffer)


class TestHGFTransformer(ut.TestCase):
    r"""Test cases for FingerprintsTransformer"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
    ]

    def test_hgf_pickling(self):
        transf = PretrainedHFTransformer(dtype=np.float32, pooling="sum")
        with tempfile.NamedTemporaryFile(delete=True) as pickled_file:
            joblib.dump(transf, pickled_file.name)
            trans2 = joblib.load(pickled_file.name)
        ori_feat = transf(dm.freesolv().smiles.values[:10])
        reloaded_feat = trans2(dm.freesolv().smiles.values[:10])
        np.testing.assert_array_equal(ori_feat, reloaded_feat)

    def test_hgf_pretrained(self):
        transf = PretrainedHFTransformer(dtype=np.float32, pooling="sum")
        fps = transf(self.smiles, enforce_dtype=True)
        fps2 = transf(self.smiles, enforce_dtype=True)
        self.assertEqual(len(fps), 3)
        self.assertEqual(len(transf), fps[0].shape[-1])
        np.testing.assert_array_equal(fps, fps2)

    @pytest.mark.xfail(reason="Cache might not be faster")
    def test_hgf_pretrained_cache(self):
        transf = PretrainedHFTransformer(dtype=np.float32, pooling="mean", precompute_cache=True)
        t0 = time.time()
        time_buffer = 1
        fps = transf.transform(self.smiles)
        ori_run = time.time() - t0
        fps2 = transf.transform(self.smiles)
        cached_run = time.time() - t0 - ori_run
        np.testing.assert_array_equal(fps, fps2)
        # add buffers
        self.assertLessEqual(cached_run, ori_run + time_buffer)


if __name__ == "__main__":
    ut.main()



from typing import Dict, Iterable, List
from typing import Any
from typing import Union
from typing import Optional
from typing import Callable
from typing import Mapping

import ast
import abc
import atexit
import copy
import glob
import pathlib
import uuid
import shelve
import platformdirs
import h5py
import os
import fsspec
import numpy as np
import pandas as pd
import datamol as dm
import joblib
import itertools
import random
import multiprocessing as mp

import pandas.errors

from functools import partial
from rdkit.Chem import rdchem
from molfeat.utils import commons
from molfeat.utils import datatype


class MolToKey:
    """Convert a molecule to a key"""

    SUPPORTED_HASH_FN = {
        "dm.unique_id": dm.unique_id,
        "dm.to_inchikey": dm.to_inchikey,
    }

    def __init__(self, hash_fn: Optional[Union[Callable, str]] = "dm.unique_id"):
        """Init function for molecular key generator.

        Args:
            hash_fn: hash function to use for the molecular key
        """

        if isinstance(hash_fn, str):
            if hash_fn not in self.SUPPORTED_HASH_FN:
                raise ValueError(
                    f"Hash function {hash_fn} is not supported. "
                    f"Supported hash functions are: {self.SUPPORTED_HASH_FN.keys()}"
                )

            self.hash_name = hash_fn
            self.hash_fn = self.SUPPORTED_HASH_FN[hash_fn]

        else:
            self.hash_fn = hash_fn
            self.hash_name = None

            if self.hash_fn is None:
                self.hash_fn = dm.unique_id
                self.hash_name = "dm.unique_id"

    def __call__(self, mol: rdchem.Mol):
        """Convert a molecule object to a key that can be used for the cache system

        Args:
            mol: input molecule object
        """
        is_mol = dm.to_mol(mol) is not None

        if is_mol and self.hash_fn is not None:
            return self.hash_fn(mol)

        return mol

    def to_state_dict(self):
        """Serialize MolToKey to a state dict."""

        if self.hash_name is None:
            raise ValueError(
                "The hash function has been provided as a function and not a string. "
                "So it's impossible to save the state. You must specifiy the hash function as a string instead."
            )

        state = {}
        state["hash_name"] = self.hash_name
        return state

    @staticmethod
    def from_state_dict(state: dict) -> "MolToKey":
        """Load a MolToKey object from a state dict."""
        return MolToKey(hash_fn=state["hash_name"])


class _Cache(abc.ABC):

    """Implementation of a cache interface"""

    def __init__(
        self,
        mol_hasher: Optional[Union[Callable, str, MolToKey]] = None,
        name: Optional[str] = None,
        n_jobs: Optional[int] = -1,
        verbose: Union[bool, int] = False,
    ):
        """
        Constructor for the Cache system

        Args:
            mol_hasher: function to use to hash molecules. If not provided, inchikey is used by default
            name: name of the cache, will be autogenerated if not provided
            n_jobs: number of parallel jobs to use when performing any computation
            verbose: whether to print progress. Default to False
        """
        self.name = name or str(uuid.uuid4())
        self.n_jobs = n_jobs
        self.verbose = verbose

        if isinstance(mol_hasher, MolToKey):
            self.mol_hasher = mol_hasher
        else:
            self.mol_hasher = MolToKey(mol_hasher)

        self.cache = {}

    def __getitem__(self, key):
        key = self.mol_hasher(key)
        return self.cache[key]

    def __contains__(self, key: Any):
        """Check whether a key is in the cache
        Args:
            key: key to check in the cache
        """
        key = self.mol_hasher(key)
        return key in self.cache

    def __len__(self):
        """Return the length of the cache"""
        return len(self.keys())

    def __iter__(self):
        """Iterate over the cache"""
        return iter(self.cache)

    def __setitem__(self, key: Any, item: Any):
        """Add an item to the cache

        Args:
            key: input key to set
            item: value of the key to set
        """
        self.update({key: item})

    def __call__(
        self,
        mols: List[Union[rdchem.Mol, str]],
        featurizer: Any,
        enforce_dtype=True,
        **transform_kwargs,
    ):
        """
        Compute the features for a list of molecules and save them to the cache

        Args:
            mols: list of molecule to preprocess
            featurizer: input featurizer to use to compute the molecular representation. Should implement a `__call__` method.
            transformer_kwargs: keyword arguments to pass to the transformer.

        !!! note
            Parquet format does not support tensor datatype, so you should ensure that the output of
            the featurizer is a numpy array in that case

        Returns:
            processed: list of computed features for input molecules
        """

        converter = copy.deepcopy(self.mol_hasher)
        mol_ids = dm.parallelized(
            converter,
            mols,
            n_jobs=self.n_jobs,
            progress=self.verbose,
            tqdm_kwargs=dict(leave=False),
        )

        # only recompute on unseen ids
        unseen_ids = []
        mol_queries = []
        for mol_id, m in zip(mol_ids, mols):
            if mol_id not in self:
                unseen_ids.append(mol_id)
                mol_queries.append(m)
        if len(mol_queries) > 0:
            features = featurizer(
                mol_queries,
                **transform_kwargs,
            )
            dtype = getattr(featurizer, "dtype", None)
            if dtype is not None:
                features = datatype.cast(features, dtype=dtype)
            for key, feat in zip(unseen_ids, features):
                self[key] = feat
            self._sync_cache()
        return self.fetch(mols)

    def clear(self, *args, **kwargs):
        ...

    @abc.abstractmethod
    def update(self, new_cache: Mapping[Any, Any]):
        ...

    def get(self, key, default: Optional[Any] = None):
        """Get the cached value for a specific key
        Args:
            key: key to get
            default: default value to return when the key is not found
        """
        key = self.mol_hasher(key)
        return self.cache.get(key, default)

    def keys(self):
        """Get list of keys in the cache"""
        return self.cache.keys()

    def values(self):
        """Get list of values in the cache"""
        return self.cache.values()

    def items(self):
        """Return iterator of key, values in the cache"""
        return self.cache.items()

    def to_dict(self):
        """Convert current cache to a dictionary"""
        return dict(self.items())

    def _sync_cache(self):
        ...

    def fetch(
        self,
        mols: List[Union[rdchem.Mol, str]],
    ):
        """Get the representation for a single

        Args:
            mols: list of molecules
        """
        if isinstance(mols, str) or not isinstance(mols, Iterable):
            mols = [mols]

        converter = copy.deepcopy(self.mol_hasher)
        mol_ids = dm.parallelized(
            converter,
            mols,
            n_jobs=self.n_jobs,
            progress=self.verbose,
            tqdm_kwargs=dict(leave=False),
        )
        return [self.get(mol_id) for mol_id in mol_ids]

    @abc.abstractclassmethod
    def load_from_file(cls, filepath: Union[os.PathLike, str], **kwargs):
        """Load a cache from a file (including remote file)

        Args:
            filepath: path to the file to load
            kwargs: keyword arguments to pass to the constructor
        """
        ...

    @abc.abstractmethod
    def save_to_file(self, filepath: Union[os.PathLike, str]):
        """Save the cache to a file

        Args:
            filepath: path to the file to save
        """
        ...


class DataCache(_Cache):
    """
    Molecular features caching system that cache computed values in memory for reuse later
    """

    def __init__(
        self,
        name: str,
        n_jobs: int = -1,
        mol_hasher: Optional[Union[Callable, str, MolToKey]] = None,
        verbose: Union[bool, int] = False,
        cache_file: Optional[Union[os.PathLike, bool]] = None,
        delete_on_exit: bool = False,
        clear_on_exit: bool = True,
    ):
        """Precomputed fingerprint caching callback

        Args:
            name: name of the cache
            n_jobs: number of parallel jobs to use when performing any computation
            mol_hasher: function to use to hash molecules. If not provided, `dm.unique_id`` is used by default
            verbose: whether to print progress. Default to False
            cache_file: Cache location. Defaults to None, which will use in-memory caching.
            delete_on_exit: Whether to delete the cache file on exit. Defaults to False.
            clear_on_exit: Whether to clear the cache on exit of the interpreter. Default to True
        """
        super().__init__(name=name, mol_hasher=mol_hasher, n_jobs=n_jobs, verbose=verbose)

        if cache_file is True:
            cache_file = pathlib.Path(
                platformdirs.user_cache_dir(appname="molfeat")
            ) / "precomputed/{}_{}.db".format(self.name, str(uuid.uuid4())[:8])

            cache_file = str(cache_file)
        self.cache_file = cache_file
        self.cache = {}
        self._initialize_cache()
        self.delete_on_exit = delete_on_exit
        self.clear_on_exit = clear_on_exit
        if self.clear_on_exit:
            atexit.register(partial(self.clear, delete=delete_on_exit))

    def _initialize_cache(self):
        if self.cache_file not in [None, False]:
            # force creation of cache directory
            cache_parent = pathlib.Path(self.cache_file).parent
            cache_parent.mkdir(parents=True, exist_ok=True)
            self.cache = shelve.open(self.cache_file)
        else:
            self.cache = {}

    def clear(self, delete: bool = False):
        """Clear cache memory if needed.
        Note that a cleared cache cannot be used anymore

        Args:
            delete: whether to delete the cache file if on disk
        """
        self.cache.clear()
        if isinstance(self.cache, shelve.Shelf):
            self.cache.close()
            # EN: temporary set it to a dict before reopening
            # this needs to be done to prevent operating on close files
            self.cache = {}
        if delete:
            if self.cache_file is not None:
                for path in glob.glob(str(self.cache_file) + "*"):
                    try:
                        os.unlink(path)
                    except:
                        pass
        else:
            self._initialize_cache()

    def update(self, new_cache: Mapping[Any, Any]):
        """Update the cache with new values

        Args:
            new_cache: new cache with items to use to update current cache
        """
        for k, v in new_cache.items():
            k = self.mol_hasher(k)
            self.cache[k] = v
        return self

    def _sync_cache(self):
        """Perform a cache sync to ensure values are up to date"""
        if isinstance(self.cache, shelve.Shelf):
            self.cache.sync()

    @classmethod
    def load_from_file(cls, filepath: Union[os.PathLike, str]):
        """Load a datache from a file (including remote file)

        Args:
            filepath: path to the file to load
        """
        cached_data = None
        with fsspec.open(filepath, "rb") as f:
            cached_data = joblib.load(f)
        data = cached_data.pop("data", {})
        new_cache = cls(**cached_data)
        new_cache.update(data)
        return new_cache

    def save_to_file(self, filepath: Union[os.PathLike, str]):
        """Save the cache to a file

        Args:
            filepath: path to the file to save
        """
        information = dict(
            name=self.name,
            n_jobs=self.n_jobs,
            mol_hasher=self.mol_hasher,
            verbose=self.verbose,
            cache_file=(self.cache_file is not None),
            delete_on_exit=self.delete_on_exit,
        )
        information["data"] = self.to_dict()
        with fsspec.open(filepath, "wb") as f:
            joblib.dump(information, f)


class MPDataCache(DataCache):
    """A datacache that supports multiprocessing natively"""

    def __init__(
        self,
        name: Optional[str] = None,
        n_jobs: int = -1,
        mol_hasher: Optional[Union[Callable, str, MolToKey]] = None,
        verbose: Union[bool, int] = False,
        clear_on_exit: bool = False,
    ):
        """Multiprocessing datacache that save cache into a shared memory

        Args:
            name: name of the cache
            n_jobs: number of parallel jobs to use when performing any computation
            mol_hasher: function to use to hash molecules. If not provided, `dm.unique_id`` is used by default
            verbose: whether to print progress. Default to False
            clear_on_exit: Whether to clear the cache on exit. Default is False to allow sharing the cache content
        """
        super().__init__(
            name=name,
            n_jobs=n_jobs,
            mol_hasher=mol_hasher,
            cache_file=None,
            verbose=verbose,
            delete_on_exit=False,
            clear_on_exit=clear_on_exit,
        )

    def _initialize_cache(self):
        """Initialize empty cache using a shared dict"""
        manager = mp.Manager()  # this might not be a great idea to initialize everytime...
        self.cache = manager.dict()


class FileCache(_Cache):
    """
    Read only cache that holds in precomputed data in a pickle, csv or h5py file.

    The convention used requires the 'keys' and  'values' columns when
    the input file needs to be loaded as a dataframe.
    """

    _PICKLE_PROTOCOL = 4
    SUPPORTED_TYPES = ["pickle", "pkl", "csv", "parquet", "pq", "hdf5", "h5"]

    def __init__(
        self,
        cache_file: Union[os.PathLike, str],
        name: Optional[str] = None,
        mol_hasher: Optional[Union[Callable, str, MolToKey]] = None,
        n_jobs: Optional[int] = None,
        verbose: Union[bool, int] = False,
        file_type: str = "parquet",
        clear_on_exit: bool = True,
        parquet_kwargs: Optional[Dict[Any, Any]] = None,
    ):
        """Precomputed fingerprint caching callback

        !!! note
            Do not pickle this object, instead use the provided saving methods.

        Args:
            cache_file: Cache location. Can be a local file or a remote file
            name: optional name to give the cache
            mol_hasher: function to use to hash molecules. If not provided, `dm.unique_id` is used by default
            n_jobs: number of parallel jobs to use when performing any computation
            verbose: whether to print information about the cache
            clear_on_exit: whether to clear the cache on exit of the interpreter
            file_type: File type that was provided. One of "csv", "pickle", "hdf5" and "parquet"
                For "csv" and "parquet", we expect columns "keys" and "values"
                For a pickle, we expect either a mapping or a dataframe with "keys" and "values" columns
            parquet_kwargs: Argument to pass to the parquet reader.
        """
        super().__init__(name=name, mol_hasher=mol_hasher, n_jobs=n_jobs, verbose=verbose)

        self.cache_file = cache_file
        self.file_type = file_type
        self.parquet_kwargs = parquet_kwargs or {}
        self.clear_on_exit = clear_on_exit

        if self.file_type not in FileCache.SUPPORTED_TYPES:
            raise ValueError(
                f"Unsupported file type, expected one of {FileCache.SUPPORTED_TYPES}, got '{self.file_type}'"
            )

        if self.cache_file is not None and dm.fs.exists(self.cache_file):
            self._load_cache()
        else:
            self.cache = {}

        if self.clear_on_exit:
            atexit.register(self.clear)

    def clear(self):
        """Clear cache memory at exit and close any open file
        Note that a cleared cache cannot be used anymore !
        """
        if self.file_type in ["hdf5", "h5"]:
            self.cache.close()
        else:
            del self.cache
        # reset cache to empty
        self.cache = {}

    def items(self):
        """Return iterator of key, values in the cache"""
        if self.file_type in ["hdf5", "h5"]:
            return ((k, np.asarray(v)) for k, v in self.cache.items())
        return super().items()

    def _load_cache(self):
        """Load cache internally if needed"""

        file_exists = dm.utils.fs.exists(self.cache_file)

        if self.file_type in ["hdf5", "h5"]:
            f = fsspec.open("simplecache::" + self.cache_file, "rb+").open()
            self.cache = h5py.File(f, "r+")

        elif not file_exists:
            self.cache = {}

        elif self.file_type in ["pickle", "pkl"]:
            with fsspec.open(self.cache_file, "rb") as IN:
                self.cache = joblib.load(IN)

        elif self.file_type == "csv":
            with fsspec.open(self.cache_file, "rb") as IN:
                # Allow the CSV file to exist but with an empty content
                try:
                    self.cache = pd.read_csv(
                        IN,
                        converters={"values": lambda x: commons.unpack_bits(ast.literal_eval(x))},
                    )
                except pandas.errors.EmptyDataError:
                    self.cache = {}

        elif self.file_type in ["parquet", "pq"]:
            self.cache = pd.read_parquet(
                self.cache_file,
                columns=["keys", "values"],
                **self.parquet_kwargs,
            )
        # convert dataframe to dict if needed
        if isinstance(self.cache, pd.DataFrame):
            self.cache = self.cache.set_index("keys").to_dict()["values"]

    def update(self, new_cache: Mapping[Any, Any]):
        """Update the cache with new values

        Args:
            new_cache: new cache with items to use to update current cache
        """
        for k, v in new_cache.items():
            key = self.mol_hasher(k)
            if self.file_type in ["hdf5", "h5"]:
                self.cache.create_dataset(key, data=v)
            else:
                self.cache[key] = v
        return self

    @classmethod
    def load_from_file(cls, filepath: Union[os.PathLike, str], **kwargs):
        """Load a FileCache from a file

        Args:
            filepath: path to the file to load
            kwargs: keyword arguments to pass to the constructor
        """
        new_cache = cls(cache_file=filepath, **kwargs)
        return new_cache

    def to_dataframe(self, pack_bits: bool = False):
        """Convert the cache to a dataframe. The converted dataframe would have `keys` and `values` columns

        Args:
            pack_bits: whether to pack the values columns into bits.
                By using molfeat.utils.commons.unpack_bits, the values column can be reloaded as an array
        """
        if pack_bits:
            loaded_items = [
                (k, commons.pack_bits(x, protocol=self._PICKLE_PROTOCOL)) for k, x in self.items()
            ]
        else:
            loaded_items = self.items()
        df = pd.DataFrame(loaded_items, columns=["keys", "values"])
        return df

    def save_to_file(
        self,
        filepath: Optional[Union[os.PathLike, str]] = None,
        file_type: Optional[str] = None,
        **kwargs,
    ):
        """Save the cache to a file

        Args:
            filepath: path to the file to save. If None, the cache is saved to the original file.
            file_type: format used to save the cache to file one of "pickle", "csv", "hdf5", "parquet".
                If None, the original file type is used.
            kwargs: keyword arguments to pass to the serializer to disk (e.g to pass to pd.to_csv or pd.to_parquet)
        """

        if filepath is None:
            filepath = self.cache_file

        if file_type is None:
            file_type = self.file_type

        if file_type in ["pkl", "pickle"]:
            with fsspec.open(filepath, "wb") as f:
                joblib.dump(self.to_dict(), f)

        elif file_type in ["csv", "parquet", "pq"]:
            df = self.to_dataframe(pack_bits=(file_type == "csv"))

            if file_type == "csv":
                with fsspec.open(filepath, "w") as f:
                    df.to_csv(f, index=False, **kwargs)
            else:
                df.to_parquet(filepath, index=False, **kwargs)

        elif file_type in ["hdf5", "h5"]:
            with fsspec.open(filepath, "wb") as IN:
                with h5py.File(IN, "w") as f:
                    for k, v in self.items():
                        f.create_dataset(k, data=v)
        else:
            raise ValueError("Unsupported output protocol: {}".format(file_type))

    def to_state_dict(self, save_to_file: bool = True) -> dict:
        """Serialize the cache to a state dict.

        Args:
            save_to_file: whether to save the cache to file.
        """

        if save_to_file is True:
            self.save_to_file()

        state = {}
        state["_cache_name"] = "FileCache"
        state["cache_file"] = self.cache_file
        state["name"] = self.name
        state["n_jobs"] = self.n_jobs
        state["verbose"] = self.verbose
        state["file_type"] = self.file_type
        state["clear_on_exit"] = self.clear_on_exit
        state["parquet_kwargs"] = self.parquet_kwargs
        state["mol_hasher"] = self.mol_hasher.to_state_dict()

        return state

    @staticmethod
    def from_state_dict(state: dict, override_args: Optional[dict] = None) -> "FileCache":
        # Don't alter the original state dict
        state = copy.deepcopy(state)

        cache_name = state.pop("_cache_name")

        if cache_name != "FileCache":
            raise ValueError(f"The cache object name is invalid: {cache_name}")

        # Load the MolToKey object
        state["mol_hasher"] = MolToKey.from_state_dict(state["mol_hasher"])

        if override_args is not None:
            state.update(override_args)

        return FileCache(**state)


class CacheList:
    """Proxy for supporting search using a list of cache"""

    def __init__(self, *caches):
        self.caches = caches

    def __getitem__(self, key):
        for cache in self.caches:
            val = cache.get(key)
            if val is not None:
                return val
        raise KeyError(f"{key} not found in any cache")

    def __contains__(self, key: Any):
        """Check whether a key is in the cache
        Args:
            key: key to check in the cache
        """
        return any(key in cache for cache in self.caches)

    def __len__(self):
        """Return the length of the cache"""
        return sum(len(c) for c in self.caches)

    def __iter__(self):
        """Iterate over all the caches"""
        return itertools.chain(*iter(self.cache))

    def __setitem__(self, key: Any, item: Any):
        """Add an item to the cache

        Args:
            key: input key to set
            item: value of the key to set
        """
        # select a random cache and add the item to the cache
        cache = random.choice(self.caches)
        cache.update({key: item})

    def __call__(self, *args, **kwargs):
        """
        Compute the features for a list of molecules and save them to the cache
        """

        raise NotImplementedError(
            "Dynamic updating of a cache list using a featurizer is not supported!"
        )

    def clear(self, *args, **kwargs):
        """Clear all the caches and make them inaccesible"""
        for cache in self.caches:
            cache.clear(*args, **kwargs)

    def update(self, new_cache: Mapping[Any, Any]):
        cache = random.choice(self.caches)
        cache.update(new_cache)

    def get(self, key, default: Optional[Any] = None):
        """Get the cached value for a specific key
        Args:
            key: key to get
            default: default value to return when the key is not found
        """
        for cache in self.caches:
            val = cache.get(key)
            if val is not None:
                return val
        return default

    def keys(self):
        """Get list of keys in the cache"""
        return list(itertools.chain(*(c.keys() for c in self.caches)))

    def values(self):
        """Get list of values in the cache"""
        return list(itertools.chain(*(c.values() for c in self.caches)))

    def items(self):
        """Return iterator of key, values in the cache"""
        return list(itertools.chain(*(c.items() for c in self.caches)))

    def to_dict(self):
        """Convert current cache to a dictionary"""
        return dict(self.items())

    def fetch(
        self,
        mols: List[Union[rdchem.Mol, str]],
    ):
        """Get the representation for a single

        Args:
            mols: list of molecules
        """
        if isinstance(mols, str) or not isinstance(mols, Iterable):
            mols = [mols]
        return [self.get(mol) for mol in mols]



#!
from typing import Union
from typing import Optional

from functools import partial

import copy
import datamol as dm
from rdkit.Avalon import pyAvalonTools
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import rdReducedGraphs
from rdkit.Chem import rdmolops
from rdkit.Chem import rdchem
from rdkit.Chem.EState import Fingerprinter as EStateFingerprinter

from loguru import logger
from molfeat.calc._mhfp import SECFP
from molfeat.calc._map4 import MAP4
from molfeat.calc.base import SerializableCalculator
from molfeat.utils.datatype import to_numpy, to_fp
from molfeat.utils.commons import fold_count_fp


FP_FUNCS = {
    "maccs": rdMolDescriptors.GetMACCSKeysFingerprint,
    "avalon": pyAvalonTools.GetAvalonFP,
    "ecfp": rdMolDescriptors.GetMorganFingerprintAsBitVect,
    "fcfp": partial(rdMolDescriptors.GetMorganFingerprintAsBitVect, useFeatures=True),
    "topological": rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect,
    "atompair": rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect,
    "rdkit": rdmolops.RDKFingerprint,
    "pattern": rdmolops.PatternFingerprint,
    "layered": rdmolops.LayeredFingerprint,
    "map4": MAP4,
    "secfp": SECFP,
    "erg": rdReducedGraphs.GetErGFingerprint,
    "estate": lambda x, **params: EStateFingerprinter.FingerprintMol(x)[0],
    "avalon-count": pyAvalonTools.GetAvalonCountFP,
    "rdkit-count": rdmolops.UnfoldedRDKFingerprintCountBased,
    "ecfp-count": rdMolDescriptors.GetHashedMorganFingerprint,
    "fcfp-count": rdMolDescriptors.GetHashedMorganFingerprint,
    "topological-count": rdMolDescriptors.GetHashedTopologicalTorsionFingerprint,
    "atompair-count": rdMolDescriptors.GetHashedAtomPairFingerprint,
}


FP_DEF_PARAMS = {
    "maccs": {},
    "avalon": {
        "nBits": 512,
        "isQuery": False,
        "resetVect": False,
        "bitFlags": pyAvalonTools.avalonSimilarityBits,
    },
    "ecfp": {
        "radius": 2,  # ECFP4
        "nBits": 2048,
        "invariants": [],
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": False,
    },
    "fcfp": {
        "radius": 2,  # FCFP4
        "nBits": 2048,
        "invariants": [],  # you may want to provide features invariance
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": True,
    },
    "topological": {
        "nBits": 2048,
        "targetSize": 4,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "nBitsPerEntry": 4,
        "includeChirality": False,
    },
    "atompair": {
        "nBits": 2048,
        "minLength": 1,
        "maxLength": 30,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "nBitsPerEntry": 4,
        "includeChirality": False,
        "use2D": True,
        "confId": -1,
    },
    "rdkit": {
        "minPath": 1,
        "maxPath": 7,
        "fpSize": 2048,
        "nBitsPerHash": 2,
        "useHs": True,
        "tgtDensity": 0.0,
        "minSize": 128,
        "branchedPaths": True,
        "useBondOrder": True,
        "atomInvariants": 0,
        "fromAtoms": 0,
        "atomBits": None,
        "bitInfo": None,
    },
    "pattern": {"fpSize": 2048, "atomCounts": [], "setOnlyBits": None},
    "layered": {
        "fpSize": 2048,
        "minPath": 1,
        "maxPath": 7,
        "atomCounts": [],
        "setOnlyBits": None,
        "branchedPaths": True,
        "fromAtoms": 0,
    },
    "map4": {"dimensions": 2048, "radius": 2},
    "secfp": {
        "n_permutations": 128,
        "nBits": 2048,
        "radius": 3,
        "min_radius": 1,
        "rings": True,
        "kekulize": False,
        "isomeric": False,
        "seed": 0,
    },
    "mhfp": {
        "n_permutations": 128,
        "radius": 3,
        "min_radius": 1,
        "rings": True,
        "kekulize": False,
        "isomeric": False,
        "seed": 0,
    },
    "erg": {"atomTypes": 0, "fuzzIncrement": 0.3, "minPath": 1, "maxPath": 15},
    "estate": {},
    # COUNTING FP
    "ecfp-count": {
        "radius": 2,  # ECFP4
        "nBits": 2048,
        "invariants": [],
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": False,
        "includeRedundantEnvironments": False,
    },
    "fcfp-count": {
        "radius": 2,  # FCFP4
        "nBits": 2048,
        "invariants": [],  # you may want to provide features invariance
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": True,
        "includeRedundantEnvironments": False,
    },
    "topological-count": {
        "nBits": 2048,
        "targetSize": 4,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "includeChirality": False,
    },
    "avalon-count": {
        "nBits": 512,
        "isQuery": False,
        "bitFlags": pyAvalonTools.avalonSimilarityBits,
    },
    "rdkit-count": {
        "minPath": 1,
        "maxPath": 7,
        "useHs": True,
        "branchedPaths": True,
        "useBondOrder": True,
        "atomInvariants": 0,
        "fromAtoms": 0,
        "atomBits": None,
        "bitInfo": None,
    },
    "atompair-count": {
        "nBits": 2048,
        "minLength": 1,
        "maxLength": 30,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "includeChirality": False,
        "use2D": True,
        "confId": -1,
    },
}


class FPCalculator(SerializableCalculator):
    """Fingerprint bit calculator for a molecule"""

    def __init__(
        self,
        method: str,
        length: Optional[int] = None,
        counting: bool = False,
        **method_params,
    ):
        """Compute the given fingeprint for a molecule

        !!! note
            For efficiency reason, count fingerprints are hashed and potentially
            re-folded and the count corresponds to the number of bits set to true

        Args:
            method (str): Name of the fingerprint method to use. See FPCalculator.available_fingerprints() for a list
            length (int, optional): Length of the fingerprint. Defaults to None.
                The default corresponds to the fingerpint default.
            counting (bool, optional): Whether to use the count version of the fingerprint
            method_params (dict): any parameters to the fingerprint algorithm.
                See FPCalculator.default_parameters(method) for all the parameters required by a given method.
        """
        self.method = method.lower()
        self.counting = counting or "-count" in self.method
        if self.counting and not "-count" in self.method:
            self.method = self.method + "-count"
        self.input_length = length
        if self.method not in FP_FUNCS:
            raise ValueError(f"Method {self.method} is not a supported featurizer")
        default_params = copy.deepcopy(FP_DEF_PARAMS[method])
        unknown_params = set(method_params.keys()).difference(set(default_params.keys()))
        if unknown_params:
            logger.error(f"Params: {unknown_params} are not valid for {method}")
        self.params = default_params
        self.params.update(
            {k: method_params[k] for k in method_params if k in default_params.keys()}
        )
        self._length = self._set_length(length)

    @staticmethod
    def available_fingerprints():
        """Get the list of available fingerprints"""
        return list(FP_FUNCS.keys())

    @staticmethod
    def default_parameters(method: str):
        """Get the default parameters for a given fingerprint method

        Args:
            method: name of the fingerprint method
        """
        return FP_DEF_PARAMS[method].copy()

    @property
    def columns(self):
        """
        Get the name of all the descriptors of this calculator
        """
        return [f"fp_{i}" for i in range(self._length)]

    def __len__(self):
        """Return the length of the calculator"""
        return self._length

    def _set_length(self, length=None):
        """Get the length of the featurizer"""
        fplen = length
        len_key = None
        if self.method == "maccs":
            fplen = 167
        elif self.method == "estate":
            fplen = 79
        elif self.method == "erg":
            fplen = 315
        elif self.method == "rdkit-count" and not fplen:
            fplen = 2048
        elif "nBits" in self.params.keys():
            len_key = "nBits"
            fplen = self.params[len_key]
        elif "n_permutations" in self.params.keys():
            # special case for mhfp
            len_key = "n_permutations"
            fplen = self.params[len_key]
        elif "fpSize" in self.params.keys():
            len_key = "fpSize"
            fplen = self.params[len_key]
        elif "dimensions" in self.params.keys():
            len_key = "dimensions"
            fplen = self.params[len_key]
        if len_key is not None and length:
            self.params[len_key] = length
            fplen = length
        return fplen

    def __call__(self, mol: Union[rdchem.Mol, str], raw: bool = False):
        r"""
        Compute the Fingerprint of a molecule

        Args:
            mol: the molecule of interest
            raw: whether to keep original datatype or convert to numpy. Useful for rdkit's similarity functions

        Returns:
            props (np.ndarray): list of computed rdkit molecular descriptors
        """
        mol = dm.to_mol(mol)
        fp_val = FP_FUNCS[self.method](mol, **self.params)
        if self.counting:
            fp_val = fold_count_fp(fp_val, self._length)
        if not raw:
            fp_val = to_numpy(fp_val)
        if self.counting and raw:
            # converint the counted values to SparseInt again
            fp_val = to_fp(fp_val, bitvect=False)
        return fp_val

    def __getstate__(self):
        # EN: note that the state is standardized with all the parameter
        # because of the possibility of default changing after
        state = {}
        state["length"] = self.input_length
        state["input_length"] = self.input_length
        state["method"] = self.method
        state["counting"] = self.counting
        state["params"] = self.params
        return state

    def __setstate__(self, state: dict):
        """Set the state of the featurizer"""
        self.__dict__.update(state)
        self._length = self._set_length(self.input_length)

    def to_state_dict(self):
        """Get the state dictionary"""
        state_dict = super().to_state_dict()
        cur_params = self.params
        default_params = copy.deepcopy(FP_DEF_PARAMS[state_dict["args"]["method"]])
        state_dict["args"].update(
            {
                k: cur_params[k]
                for k in cur_params
                if (cur_params[k] != default_params[k] and cur_params[k] is not None)
            }
        )
        # we want to keep all the additional parameters in the state dict
        return state_dict



from typing import Optional
from typing import Any
from typing import Union
from typing import Callable

import yaml
import joblib
import pathlib
import os
import fsspec
import tempfile
import platformdirs
import filelock
import datamol as dm
from loguru import logger

from molfeat.store.modelcard import ModelInfo
from molfeat.utils import commons


class ModelStoreError(Exception):
    pass


class ModelStore:
    """A class emulating artefact serializing from any url

    This class should not only allow pretrained model serializing and loading,
    but should also help in listing model availability and registering models

    For simplicity.
        * There is no versioning.
        * Only one model per model name is permitted
        * Model deletion is not allowed
        * Only a single Cloud storage is supported

    """

    # EN: be careful not to recreate ada
    # EN: should we just use modelstore ?
    MODEL_STORE_BUCKET = "gs://molfeat-store-prod/artifacts/"
    MODEL_PATH_NAME = "model.save"
    METADATA_PATH_NAME = "metadata.json"

    def __init__(self, model_store_bucket: Optional[str] = None):
        if model_store_bucket is None:
            model_store_bucket = self.MODEL_STORE_BUCKET
        self.model_store_bucket = model_store_bucket
        self._available_models = []
        self._update_store()

    def _update_store(self):
        """Initialize the store with all available models"""
        all_metadata = dm.fs.glob(dm.fs.join(self.model_store_bucket, "**/metadata.json"))
        self._available_models = []
        for mtd_file in all_metadata:
            with fsspec.open(mtd_file, "r") as IN:
                mtd_content = yaml.safe_load(IN)
                model_info = ModelInfo(**mtd_content)
                self._available_models.append(model_info)

    @property
    def available_models(self):
        """Return a list of all models that have been serialized in molfeat"""
        return self._available_models

    def register(
        self,
        modelcard: Union[ModelInfo, dict],
        model: Optional[Any] = None,
        chunk_size: int = 2048,
        save_fn: Optional[Callable] = None,
        save_fn_kwargs: Optional[dict] = None,
        force: bool = True,
    ):
        """
        Register a new model to the store

        Args:
            modelcard: Model information
            model: A path to the model artifact or any object that needs to be saved
            chunk_size: the chunk size for the upload
            save_fn: any custom function for serializing the model, that takes the model, the upload path and parameters as inputs
            save_fn_kwargs: any additional kwargs to pass to save_fn
            force: whether to force upload to the bucket

        """
        if not isinstance(modelcard, ModelInfo):
            modelcard = ModelInfo(**modelcard)
        # we save the model first
        if self.exists(card=modelcard):
            logger.warning(f"Model {modelcard.name} exists already ...")
            if not force:
                return

        model_root_dir = modelcard.path(self.model_store_bucket)
        model_path = model
        model_upload_path = dm.fs.join(model_root_dir, self.MODEL_PATH_NAME)
        model_metadata_upload_path = dm.fs.join(model_root_dir, self.METADATA_PATH_NAME)

        save_fn_kwargs = save_fn_kwargs or {}
        if save_fn is None:
            if not isinstance(model, (pathlib.Path, os.PathLike)):
                local_model_path = tempfile.NamedTemporaryFile(delete=False)
                with local_model_path as f:
                    joblib.dump(model, local_model_path)
                model_path = local_model_path.name
            # Upload the artifact to the bucket
            dm.fs.copy_file(
                model_path,
                model_upload_path,
                progress=True,
                leave_progress=False,
                chunk_size=chunk_size,
                force=force,
            )
        else:
            model_path = save_fn(model, model_upload_path, **save_fn_kwargs)
        modelcard.sha256sum = commons.sha256sum(model_path)
        # then we save the metadata as json
        with fsspec.open(model_metadata_upload_path, "w") as OUT:
            OUT.write(modelcard.json())
        self._update_store()
        logger.info(f"Successfuly registered model {modelcard.name} !")

    def _filelock(self, lock_name: str):
        """Create an empty lock file into `cache_dir_path/locks/lock_name`"""

        lock_path = dm.fs.join(
            str(platformdirs.user_cache_dir("molfeat")), "_lock_files", lock_name
        )
        mapper = dm.fs.get_mapper(lock_path)
        # ensure file is created
        # out = mapper.fs.touch(lock_path) # does not work  -_-
        with fsspec.open(lock_path, "w", auto_mkdir=True) as f:
            pass

        return filelock.FileLock(lock_path)

    def download(
        self,
        modelcard: ModelInfo,
        output_dir: Optional[Union[os.PathLike, pathlib.Path]] = None,
        chunk_size: int = 2048,
        force: bool = False,
    ):
        """Download an artifact locally

        Args:
            modelcard: information on the model to download
            output_dir: path where to save the downloaded artifact
            chunk_size: chunk size to use for download
            force: whether to force download even if the file exists already
        """

        remote_dir = modelcard.path(self.model_store_bucket)
        model_name = modelcard.name
        if not self.exists(modelcard, check_remote=True):
            raise ModelStoreError(f"Model {model_name} does not exist in the model store !")

        if output_dir is None:
            output_dir = dm.fs.join(platformdirs.user_cache_dir("molfeat"), model_name)

        dm.fs.mkdir(output_dir, exist_ok=True)

        model_remote_path = dm.fs.join(remote_dir, self.MODEL_PATH_NAME)
        model_dest_path = dm.fs.join(output_dir, self.MODEL_PATH_NAME)
        metadata_remote_path = dm.fs.join(remote_dir, self.METADATA_PATH_NAME)
        metadata_dest_path = dm.fs.join(output_dir, self.METADATA_PATH_NAME)

        # avoid downloading if the file exists already
        if (
            not (
                dm.fs.exists(metadata_dest_path)
                and (dm.fs.exists(model_dest_path) == dm.fs.exists(model_remote_path))
            )
            or force
        ):
            # metadata should exists if the model exists
            with self._filelock(f"{model_name}.metadata.json.lock"):
                dm.fs.copy_file(
                    metadata_remote_path,
                    metadata_dest_path,
                    progress=True,
                    leave_progress=False,
                    force=True,
                )

            if dm.fs.exists(model_remote_path):
                with self._filelock(f"{model_name}.lock"):
                    if dm.fs.is_dir(model_remote_path):
                        # we copy the model dir
                        dm.fs.copy_dir(
                            model_remote_path,
                            model_dest_path,
                            progress=True,
                            leave_progress=False,
                            chunk_size=chunk_size,
                            force=force,
                        )
                    else:
                        # we copy the model dir
                        dm.fs.copy_file(
                            model_remote_path,
                            model_dest_path,
                            progress=True,
                            leave_progress=False,
                            chunk_size=chunk_size,
                            force=force,
                        )

        cache_sha256sum = commons.sha256sum(model_dest_path)
        if modelcard.sha256sum is not None and cache_sha256sum != modelcard.sha256sum:
            mapper = dm.fs.get_mapper(output_dir)
            mapper.fs.delete(output_dir, recursive=True)
            raise ModelStoreError(
                f"""The destination artifact at {model_dest_path} has a different sha256sum ({cache_sha256sum}) """
                f"""than the Remote artifact sha256sum ({modelcard.sha256sum}). The destination artifact has been removed !"""
            )

        return output_dir

    def load(
        self,
        model_name: Union[str, dict, ModelInfo],
        load_fn: Optional[Callable] = None,
        load_fn_kwargs: Optional[dict] = None,
        download_output_dir: Optional[Union[os.PathLike, pathlib.Path]] = None,
        chunk_size: int = 2048,
        force: bool = False,
    ):
        """
        Load a model by its name

        Args:
            model_name: name of the model to load
            load_fn: Custom loading function to load the model
            load_fn_kwargs: Optional dict of additional kwargs to provide to the loading function
            download_output_dir: Argument for download function to specify the download folder
            chunk_size: chunk size for download
            force: whether to reforce the download of the file

        """
        if isinstance(model_name, str):
            # find the model with the same name
            modelcard = self.search(name=model_name)[0]
        else:
            modelcard = model_name
        output_dir = self.download(
            modelcard=modelcard,
            output_dir=download_output_dir,
            chunk_size=chunk_size,
            force=force,
        )
        if load_fn is None:
            load_fn = joblib.load
        model_path = dm.fs.join(output_dir, self.MODEL_PATH_NAME)
        metadata_path = dm.fs.join(output_dir, self.METADATA_PATH_NAME)

        # deal with non-pretrained models that might not have a serialized file
        model = None
        load_fn_kwargs = load_fn_kwargs or {}
        if dm.fs.exists(model_path):
            model = load_fn(model_path, **load_fn_kwargs)
        with fsspec.open(metadata_path, "r") as IN:
            model_info_dict = yaml.safe_load(IN)
        model_info = ModelInfo(**model_info_dict)
        return model, model_info

    def __contains__(self, card: Optional[ModelInfo] = None):
        return self.exists(card)

    def exists(
        self,
        card: Optional[ModelInfo] = None,
        check_remote: bool = False,
        **match_params,
    ) -> bool:
        """Returns True if a model is registered in the store

        Args:
            card: card of the model to check
            check_remote: whether to check if the remote path of the model exists
            match_params: parameters for matching as expected by `ModelInfo.match`
        """

        found = False
        for model_info in self.available_models:
            if model_info.match(card, **match_params):
                found = True
                break
        return found and (not check_remote or dm.fs.exists(card.path(self.model_store_bucket)))

    def search(self, modelcard: Optional[ModelInfo] = None, **search_kwargs):
        """ "Return all model card that match the required search parameters

        Args:
            modelcard: model card to search for
            search_kwargs: search parameters to use
        """
        search_infos = {}
        found = []
        if modelcard is not None:
            search_infos = modelcard.dict().copy()
        search_infos.update(search_kwargs)
        for model in self.available_models:
            if model.match(search_infos, match_only=list(search_infos.keys())):
                found.append(model)
        return found



from typing import Optional
from typing import Union
from typing import Callable
from typing import List
from collections.abc import Iterable
from functools import lru_cache

import copy
import datamol as dm

from loguru import logger
from rdkit.Chem import rdchem
from molfeat.utils.commons import _parse_to_evaluable_str
from molfeat.trans.base import MoleculeTransformer
from molfeat.utils.cache import DataCache
from molfeat.store.loader import PretrainedModel


class PretrainedMolTransformer(MoleculeTransformer):
    r"""
    Transformer based on pretrained featurizer

    !!! note
        * When implementing a subclass of this class, you need to define the `_embed` and optionally the `_convert` methods.
        * If your model is an instance of PretrainedModel that handles loading of the model from a store or through a complex mechanism
          then you can decide whether you want to preload the true underlying model. You will be in charge of handling the logic of when you need to call preload, and when you don't.
          Note however that by default preloading is only attempted when the featurizer is still an instance of PretrainedModel.


    Attributes
        featurizer (object): featurizer object
        dtype (type, optional): Data type. Use call instead
        precompute_cache: (bool, optional): Whether to precompute the features into a local cache. Defaults to False.
            Note that due to molecular hashing, some pretrained featurizers might be better off just not using any cache as they can be faster.
            Furthermore, the cache is not saved when pickling the object. If you want to save the cache, you need to save the object separately.
        _require_mols (bool): Whether the embedding takes mols or smiles as input
        preload: whether to preload the pretrained model from the store (if available) during initialization.

    """

    def __init__(
        self,
        dtype: Optional[Callable] = None,
        precompute_cache: Optional[Union[bool, DataCache]] = None,
        preload: bool = False,
        **params,
    ):
        self._save_input_args()

        featurizer = params.pop("featurizer", None)
        super().__init__(dtype=dtype, featurizer="none", **params)
        self.featurizer = featurizer
        self._require_mols = False
        self.preload = preload
        self._feat_length = None
        if precompute_cache is False:
            precompute_cache = None
        if precompute_cache is True:
            name = str(self.__class__.__name__)
            precompute_cache = DataCache(name=name)
        self.precompute_cache = precompute_cache

    def set_cache(self, cache: DataCache):
        """Set the cache for the transformer

        Args:
            cache: cache object
        """
        self.precompute_cache = cache

    def _get_param_names(self):
        """Get parameter names for the estimator"""
        out = self._input_params.keys()
        out = [x for x in out if x != "featurizer"]
        return out

    def _embed(self, smiles: str, **kwargs):
        """Internal molecular embedding
        This functiom takes a list of smiles or molecules and return the featurization
        corresponding to the inputs.  In `transform` and `_transform`, this function is
        called after calling `_convert`

        Args:
            smiles: input smiless
        """
        raise NotImplementedError

    def _preload(self):
        """Preload the pretrained model for later queries"""
        if self.featurizer is not None and isinstance(self.featurizer, PretrainedModel):
            self.featurizer = self.featurizer.load()
            self.preload = True

    def __getstate__(self):
        """Getting state to allow pickling"""
        d = copy.deepcopy(self.__dict__)
        d["precompute_cache"] = None
        if isinstance(getattr(self, "featurizer", None), PretrainedModel) or self.preload:
            d.pop("featurizer", None)
        return d

    def __setstate__(self, d):
        """Setting state during reloading pickling"""
        self.__dict__.update(d)
        self._update_params()

    def fit(self, *args, **kwargs):
        return self

    def _convert(self, inputs: list, **kwargs):
        """Convert molecules to the right format

        In `transform` and `_transform`, this function is called before calling `_embed`

        Args:
            inputs: inputs to preprocess

        Returns:
            processed: pre-processed input list
        """
        if not self._require_mols:
            inputs = [dm.to_smiles(m) for m in inputs]
        return inputs

    def preprocess(self, inputs: list, labels: Optional[list] = None):
        """Run preprocessing on the input data
        Args:
            inputs: list of input molecules
            labels: list of labels
        """
        out = super().preprocess(inputs, labels)
        if self.precompute_cache not in [False, None]:
            try:
                self.transform(inputs)
            except:
                pass
        return out

    def _transform(self, mol: rdchem.Mol, **kwargs):
        r"""
        Compute features for a single molecule.
        This method would potentially need to be reimplemented by any child class

        Args:
            mol (rdchem.Mol): molecule to transform into features

        Returns
            feat: featurized input molecule

        """
        feat = None
        if self.precompute_cache is not None:
            feat = self.precompute_cache.get(mol)
        if feat is None:
            try:
                mols = [dm.to_mol(mol)]
                mols = self._convert(mols, **kwargs)
                feat = self._embed(mols, **kwargs)
                feat = feat[0]
            except Exception as e:
                if self.verbose:
                    logger.error(e)

            if self.precompute_cache is not None:
                self.precompute_cache[mol] = feat
        return feat

    def transform(self, smiles: List[str], **kwargs):
        """Perform featurization of the input molecules

        The featurization process is as follow:
        1. convert the input molecules into the right format, expected by the pre-trained model using `_convert`
        2. compute embedding of the molecule using `_embed`
        3. perform any model-specific postprocessing and cache update

        The dtype returned is the native datatype of the transformer.
        Use `__call__` to get the dtype in the `dtype` attribute format

        Args:
            mols: a list containing smiles or mol objects

        Returns:
            out: featurized molecules
        """
        if isinstance(smiles, str) or not isinstance(smiles, Iterable):
            smiles = [smiles]

        n_mols = len(smiles)
        ind_to_compute = dict(zip(range(n_mols), range(n_mols)))
        pre_computed = [None] * n_mols

        if self.precompute_cache not in [False, None]:
            ind_to_compute = {}
            pre_computed = self.precompute_cache.fetch(smiles)
            ind = 0
            for i, v in enumerate(pre_computed):
                if v is None:
                    ind_to_compute[i] = ind
                    ind += 1

        parallel_kwargs = getattr(self, "parallel_kwargs", {})
        mols = dm.parallelized(
            dm.to_mol, smiles, n_jobs=getattr(self, "n_jobs", 1), **parallel_kwargs
        )
        mols = [mols[i] for i in ind_to_compute]

        if len(mols) > 0:
            converted_mols = self._convert(mols, **kwargs)
            out = self._embed(converted_mols, **kwargs)

            if not isinstance(out, list):
                out = list(out)

            if self.precompute_cache is not None:
                # cache value now
                self.precompute_cache.update(dict(zip(mols, out)))
        out = [
            out[ind_to_compute[i]] if i in ind_to_compute else pre_computed[i]
            for i in range(n_mols)
        ]
        return out

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return str(self) == str(other)
        return False

    def _update_params(self):
        self._fitted = False

    def __len__(self):
        if self._feat_length is None:
            self._preload()
            tmp_mol = dm.to_mol("CCC")
            embs = self._transform(tmp_mol)
            self._feat_length = len(embs)
        return self._feat_length

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return "{}(dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.dtype),
        )



from typing import Callable
from typing import List
from typing import Optional

import numpy as np
import torch
import copy
from molfeat.utils.commons import _parse_to_evaluable_str
from molfeat.utils import requires
from molfeat.utils.pooler import Pooling
from molfeat.trans.pretrained.base import PretrainedMolTransformer

if requires.check("graphormer_pretrained"):
    from graphormer_pretrained.embeddings import GraphormerEmbeddingsExtractor
    from graphormer_pretrained.data.smiles_dataset import GraphormerInferenceDataset


class GraphormerTransformer(PretrainedMolTransformer):
    r"""
    Graphormer transformer based on pretrained sequence embedder

    Attributes:
        featurizer: Graphormer embedding object
        dtype: Data type. Use call instead
        pooling: Pooling method for Graphormer's embedding layer
    """

    def __init__(
        self,
        kind: str = "pcqm4mv2_graphormer_base",
        dtype: Callable = np.float32,
        pooling: str = "mean",
        max_length: Optional[int] = None,
        version=None,
        **params,
    ):
        super().__init__(dtype=dtype, pooling=pooling, **params)
        if not requires.check("graphormer_pretrained"):
            raise ValueError("`graphormer` is required to use this featurizer.")

        self.preload = True
        self.name = kind
        self._require_mols = False
        self.max_length = max_length
        self.pooling = pooling
        if isinstance(pooling, str):
            pooling = Pooling(dim=1, name=pooling)
        self.pooling = pooling
        self.featurizer = GraphormerEmbeddingsExtractor(
            pretrained_name=self.name, max_nodes=self.max_length
        )
        self.featurizer.config.max_nodes = self.max_length
        self.version = version

    def __repr__(self):
        return "{}(name={}, pooling={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.name),
            _parse_to_evaluable_str(self.pooling.name),
            _parse_to_evaluable_str(self.dtype),
        )

    @staticmethod
    def list_available_models():
        """List available graphormer model to use"""
        return [
            "pcqm4mv1_graphormer_base",  # PCQM4Mv1
            "pcqm4mv2_graphormer_base",  # PCQM4Mv2
            "pcqm4mv1_graphormer_base_for_molhiv",  # ogbg-molhiv
            "oc20is2re_graphormer3d_base",  # Open Catalyst Challenge
        ]

    def _embed(self, inputs: List[str], **kwargs):
        """Internal molecular embedding

        Args:
            smiles: input smiless
        """
        with torch.no_grad():
            x = self.featurizer.model(inputs)
            x = self.pooling(x)
        return x.numpy()

    def __getstate__(self):
        """Getting state to allow pickling"""
        d = copy.deepcopy(self.__dict__)
        d["precompute_cache"] = None
        d.pop("featurizer", None)
        return d

    def __setstate__(self, d):
        """Setting state during reloading pickling"""
        self.__dict__.update(d)
        self._update_params()

    def compute_max_length(self, inputs: list):
        """Compute maximum node number for the input list of molecules

        Args:
            inputs: input list of molecules
        """
        dataset = GraphormerInferenceDataset(
            inputs,
            multi_hop_max_dist=self.featurizer.config.multi_hop_max_dist,
            spatial_pos_max=self.featurizer.config.spatial_pos_max,
        )
        xs = [item.x.size(0) for item in dataset]
        return max(xs)

    def set_max_length(self, max_length: int):
        """Set the maximum length for this featurizer"""
        self.max_length = max_length
        self._update_params()
        self._preload()

    def _convert(self, inputs: list, **kwargs):
        """Convert molecules to the right format

        Args:
            inputs: inputs to preprocess

        Returns:
            processed: pre-processed input list
        """
        inputs = super()._convert(inputs, **kwargs)
        batch = self.featurizer._convert(inputs)
        return batch

    def _update_params(self):
        super()._update_params()
        self.featurizer = GraphormerEmbeddingsExtractor(
            pretrained_name=self.name, max_nodes=self.max_length
        )
        self.featurizer.config.max_nodes = self.max_length



from typing import Optional

import abc
import os
import datamol as dm
from functools import lru_cache
from molfeat.store.modelstore import ModelStore
from molfeat.store.modelstore import ModelStoreError


class PretrainedModel(abc.ABC):
    """Base class for loading pretrained models"""

    @classmethod
    def _artifact_load(cls, name: str, download_path: Optional[os.PathLike] = None, **kwargs):
        """Load an artifact based on its name

        Args:
            name: name of the model to load
            download_path: path to a directory where to save the downloaded files
        """
        ...

    @classmethod
    def _load_or_raise(
        cls,
        name: str,
        download_path: Optional[os.PathLike] = None,
        **kwargs,
    ):
        """Load model or raise an exception

        Args:
            name: name of the model to load
            download_path: local download path of the model

        """
        ...

    @abc.abstractmethod
    def load(self):
        """Load the model"""
        ...


class PretrainedStoreModel(PretrainedModel):
    r"""
    Class for loading pretrained models from the model zoo
    """

    def __init__(
        self,
        name: str,
        cache_path: Optional[os.PathLike] = None,
        store: Optional[ModelStore] = None,
    ):
        """Interface for pretrained model from the default modelstore

        Args:
            name: name of the pretrained transformer in the model store
            cache_path: optional local cache path.
            store: ModelStore to use for loading the pretrained model
        """
        self.name = name
        self.cache_path = cache_path
        if store is None:
            store = ModelStore()
        self.store = store

    @classmethod
    def _artifact_load(cls, name: str, download_path: Optional[os.PathLike] = None, **kwargs):
        """Load internal artefact from the model store

        Args:
            name: name of the model to load
            download_path: path to a directory where to save the downloaded files
        """

        if not dm.fs.exists(download_path):
            cls._load_or_raise.cache_clear()
        return cls._load_or_raise(name, download_path, **kwargs)

    @classmethod
    @lru_cache(maxsize=100)
    def _load_or_raise(
        cls,
        name: str,
        download_path: Optional[os.PathLike] = None,
        store: Optional[ModelStore] = None,
        **kwargs,
    ):
        """Load when from ada or raise exception
        Args:
            name: name
        """
        if store is None:
            store = ModelStore()
        try:
            modelcard = store.search(name=name)[0]
            artifact_dir = store.download(modelcard, download_path, **kwargs)
        except Exception as e:
            mess = f"Can't retrieve model {name} from the store !"
            raise ModelStoreError(mess)
        return artifact_dir

    def load(self):
        """Load the model"""
        raise NotImplementedError



from molfeat.trans.pretrained.base import PretrainedMolTransformer
from molfeat.trans.pretrained.fcd import FCDTransformer
from molfeat.trans.pretrained.dgl_pretrained import PretrainedDGLTransformer
from molfeat.trans.pretrained.graphormer import GraphormerTransformer
from molfeat.trans.pretrained.hf_transformers import PretrainedHFTransformer



from typing import Union
from typing import List
from loguru import logger


class loguru_off:
    def __init__(self, modules: Union[str, List[str]]):
        """Disable loguru logging for a list of modules

        Args:
            modules: list of modules to temporary disable
        """
        if isinstance(modules, str):
            modules = [modules]
        self.modules = modules
        self._status = logger._core.enabled.copy()

    def __enter__(self):
        for mod in self.modules:
            logger.disable(mod)

    def __exit__(self, *args, **kwargs):
        for mod in self.modules:
            matching_mods = [x for x in self._status.keys() if x.startswith(mod)]
            if len(matching_mods) > 0:
                for mod_match in matching_mods:
                    logger._change_activation(mod_match, self._status[mod_match])
            else:
                # if it was not in a known state before, we will enable it
                logger.enable(mod)



import copy
from typing import Union, List, Optional
from collections import OrderedDict


import datamol as dm
import numpy as np

from loguru import logger

from rdkit.Chem.QED import properties
from rdkit.Chem import Descriptors
from rdkit.Chem import Descriptors3D
from rdkit.Chem import rdchem
from rdkit.Chem import FindMolChiralCenters
from rdkit.Chem import rdPartialCharges
from rdkit.Chem import rdMolDescriptors


from molfeat.utils.commons import requires_conformer
from molfeat.utils.commons import requires_standardization
from molfeat.utils.datatype import to_numpy
from molfeat.utils import requires
from molfeat.calc.base import SerializableCalculator

if requires.check("mordred"):
    from mordred import Calculator as MordredCalculator
    from mordred import descriptors as mordred_descriptors

else:
    MordredCalculator = requires.mock("mordred")


def _charge_descriptors_computation(mol: dm.Mol):
    """Recompute properly the RDKIT 2D Descriptors related to charge

    Args:
        mol: input molecule for which the charge descriptors should be recomputed
    """
    descrs = {}
    # we disconnect the metal from the molecule then we add the hydrogen atoms and finally
    # make sure that gasteiger is recomputed. This fixes the nan and inf issues,
    # while also making sure we are closer the the proper interpretation
    mol = dm.standardize_mol(mol, disconnect_metals=True)
    mol = dm.add_hs(mol, explicit_only=False)
    rdPartialCharges.ComputeGasteigerCharges(mol)
    atomic_charges = [float(at.GetProp("_GasteigerCharge")) for at in mol.GetAtoms()]
    atomic_charges = np.clip(atomic_charges, a_min=-500, a_max=500)
    min_charge, max_charge = np.nanmin(atomic_charges), np.nanmax(atomic_charges)
    descrs["MaxPartialCharge"] = max_charge
    descrs["MinPartialCharge"] = min_charge
    descrs["MaxAbsPartialCharge"] = max(np.abs(min_charge), np.abs(max_charge))
    descrs["MinAbsPartialCharge"] = min(np.abs(min_charge), np.abs(max_charge))
    return descrs


class RDKitDescriptors2D(SerializableCalculator):
    r"""
    Compute a list of available  rdkit 2D descriptors for a molecule.
    The descriptor calculator does not mask errors in featurization and will propagate them
    """

    DESCRIPTORS_FN = {name: fn for (name, fn) in Descriptors.descList}

    def __init__(
        self,
        replace_nan: Optional[bool] = False,
        augment: Optional[bool] = True,
        descrs: List = None,
        avg_ipc: Optional[bool] = True,
        do_not_standardize: Optional[bool] = False,
        **kwargs,
    ):
        """RDKit descriptor computation

        Args:
            replace_nan: Whether to replace nan or infinite values. Defaults to False.
            augment: Whether to augment the descriptors with some additional custom features
            descrs: Subset of available features to consider if not None
            avg_ipc: Whether to average IPC values or to use rdkit original
            do_not_standardize: Whether to force standardization of molecule before computation of the descriptor.
                Set to True if you want molfeat<=0.5.3 behaviour
        """
        self.replace_nan = replace_nan
        self.augment = augment
        self.descrs = descrs
        self.avg_ipc = avg_ipc
        self.do_not_standardize = do_not_standardize
        all_features = [d[0] for d in Descriptors.descList]
        if self.augment:
            all_features += [
                "NumAtomStereoCenters",
                "NumUnspecifiedAtomStereoCenters",
                "NumBridgeheadAtoms",
                "NumAmideBonds",
                "NumSpiroAtoms",
                "Alerts",
            ]
        if descrs is not None:
            self._columns = [x for x in descrs if x in all_features]
            unknown_descrs = set(descrs) - set(all_features)
            if len(unknown_descrs) > 0:
                logger.warning(f"Following features are not supported: {unknown_descrs}")
        else:
            self._columns = all_features

    def __getstate__(self):
        """Serialize the class for pickling."""
        state = {}
        state["replace_nan"] = self.replace_nan
        state["augment"] = self.augment
        state["descrs"] = self.descrs
        state["_columns"] = self._columns

        # EN: set `avg_ipc` and `standardize`` default value to False for compat until next release
        state["avg_ipc"] = getattr(self, "avg_ipc", False)
        state["do_not_standardize"] = getattr(self, "do_not_standardize", False)
        return state

    def _compute_extra_features(self, mol: Union[rdchem.Mol, str]):
        """Compute the extra properties required for the augmented features version

        Args:
            mol: Input molecule

        Returns:
            props (dict): Dict of extra molecular properties
        """
        mol = copy.deepcopy(mol)
        FindMolChiralCenters(mol, force=True)
        # "NumAtomStereoCenters", "NumUnspecifiedAtomStereoCenters", "NumBridgeheadAtoms", "NumAmideBonds", "NumSpiroAtoms"
        p_obj = rdMolDescriptors.Properties()
        props = OrderedDict(zip(p_obj.GetPropertyNames(), p_obj.ComputeProperties(mol)))
        # Alerts
        qed_props = properties(mol)
        props["Alerts"] = qed_props.ALERTS
        return props

    @property
    def columns(self):
        """
        Get the name of all the descriptors of this calculator
        """
        return self._columns

    def __len__(self):
        """Return the length of the calculator"""
        return len(self._columns)

    @requires_standardization(disconnect_metals=True, remove_salt=True)
    def __call__(self, mol: Union[rdchem.Mol, str]):
        r"""
        Get rdkit basic descriptors for a molecule

        Args:
            mol: the molecule of interest

        Returns:
            props (np.ndarray): list of computed rdkit molecular descriptors
        """
        mol = dm.to_mol(mol)
        vals = []
        props = {}
        if self.augment:
            props = self._compute_extra_features(mol)
        fixed_charge_descr = _charge_descriptors_computation(mol)
        for name in self.columns:
            val = float("nan")
            if name in fixed_charge_descr:
                val = fixed_charge_descr[name]
            elif name == "Ipc" and self.avg_ipc:  # bug fix of the rdkit IPC value
                val = self.DESCRIPTORS_FN[name](mol, avg=True)
            elif name in self.DESCRIPTORS_FN:
                val = self.DESCRIPTORS_FN[name](mol)
            elif name in props:
                val = props[name]
            else:
                raise ValueError(f"Property: {name} is not supported !")
            vals.append(val)
        vals = to_numpy(vals)
        if self.replace_nan:
            vals = np.nan_to_num(vals)
        return vals


class MordredDescriptors(SerializableCalculator):
    r"""
    Compute mordred descriptors.
    The descriptor calculator does not mask errors in featurization and will propagate them.

    !!! note
        Mordred descriptors can results in undefined or nan behaviour depending on your input molecule.
        It is recommended that the user handles those nan values himself by either removing the descriptor
        or imputing the missing values.
    """

    def __init__(
        self,
        ignore_3D: bool = True,
        replace_nan: bool = False,
        do_not_standardize: bool = False,
        **kwargs,
    ):
        """Mordred descriptor computation

        Args:
            ignore_3D (bool, optional): Whether to ignore 3D descriptors or include them
            replace_nan (bool, optional): Whether to replace nan or infinite values. Defaults to False.
            do_not_standardize: Whether to force standardize molecules or keep it the same

        """
        if not requires.check("mordred"):
            logger.error(
                "`mordred` is not available, please install it `pip install 'mordred[full]'`"
            )
            raise ImportError("Cannot import `mordred`")
        self.replace_nan = replace_nan
        self.ignore_3D = ignore_3D
        self.do_not_standardize = do_not_standardize
        self._calc = None
        self._init_calc()

    def _init_calc(self):
        """Initialize mordred calculator"""
        self._calc = MordredCalculator(mordred_descriptors, ignore_3D=self.ignore_3D)

    @property
    def columns(self):
        """
        Get the name of all the descriptors of this calculator
        """
        return [str(x) for x in self._calc.descriptors]

    def __len__(self):
        """Return the length of the calculator"""
        return len(self._calc)

    def __getstate__(self):
        """Serialize the class for pickling."""
        state = {}
        state["ignore_3D"] = self.ignore_3D
        state["replace_nan"] = self.replace_nan
        state["do_not_standardize"] = getattr(self, "do_not_standardize", False)
        return state

    def __setstate__(self, state: dict):
        """Reload the class from pickling."""
        self.__dict__.update(state)
        self._init_calc()

    def __call__(self, mol: Union[rdchem.Mol, str], conformer_id: Optional[int] = -1):
        r"""
        Get rdkit basic descriptors for a molecule

        Args:
            mol: the molecule of interest
            conformer_id (int, optional): Optional

        Returns:
            props (np.ndarray): list of computed mordred molecular descriptors
        """
        mol = dm.to_mol(mol)
        vals = self._calc(mol, conformer_id).fill_missing()
        vals = to_numpy(vals)
        if self.replace_nan:
            vals = np.nan_to_num(vals)
        return vals


class RDKitDescriptors3D(SerializableCalculator):
    """
    Compute a list of 3D rdkit descriptors
    """

    def __init__(
        self,
        replace_nan: bool = False,
        ignore_descrs: list = ["CalcGETAWAY"],
        **kwargs,
    ):
        """Compute 3D descriptors

        Args:
            replace_nan (bool, optional): Whether to replace nan or infinite values. Defaults to False.
            ignore_descrs (list, optional): Descriptors to ignore for performance issues. Defaults to ["CalcGETAWAY"].
        """
        self.replace_nan = replace_nan

        self._descr = [
            "CalcAsphericity",
            "CalcEccentricity",
            "CalcInertialShapeFactor",
            "CalcNPR1",
            "CalcNPR2",
            "CalcPMI1",
            "CalcPMI2",
            "CalcPMI3",
            "CalcRadiusOfGyration",
            "CalcSpherocityIndex",
            "CalcPBF",
        ]

        self.ignore_descrs = ignore_descrs or []
        self._vec_descr = [
            "CalcAUTOCORR3D",
            "CalcRDF",
            "CalcMORSE",
            "CalcWHIM",
            "CalcGETAWAY",
        ]
        self._vec_descr_length = [80, 210, 224, 114, 273]
        self._columns = [x for x in self._descr if x not in self.ignore_descrs]
        for desc, desc_len in zip(self._vec_descr, self._vec_descr_length):
            if desc in self.ignore_descrs:
                continue
            for pos in range(desc_len):
                self._columns.append(f"{desc}_{pos}")

    def __getstate__(self):
        """Serialize the class for pickling."""
        state = {}
        state["replace_nan"] = self.replace_nan
        state["ignore_descrs"] = self.ignore_descrs
        state["_columns"] = self._columns
        return state

    def __len__(self):
        """Get the length of the descriptor"""
        return len(self._columns)

    @property
    def columns(self):
        """Get the descriptors columns"""
        return self._columns

    @requires_conformer
    def __call__(self, mol: Union[rdchem.Mol, str], conformer_id: Optional[int] = -1):
        r"""
        Get rdkit 3D descriptors for a molecule

        Args:
            mol: the molecule of interest
            conformer_id (int, optional): Optional conformer id. Defaults to -1.

        Returns:
            props (np.ndarray): list of computed mordred molecular descriptors
        """

        mol = dm.to_mol(mol)
        desc_val = []
        for desc in self._descr:
            val = float("nan")
            if desc not in self.ignore_descrs:
                try:
                    val = getattr(Descriptors3D.rdMolDescriptors, desc)(mol, confId=conformer_id)
                except:
                    pass
                desc_val.append(val)
        for i, desc in enumerate(self._vec_descr):
            val = [float("nan")] * self._vec_descr_length[i]
            if desc not in self.ignore_descrs:
                try:
                    val = getattr(Descriptors3D.rdMolDescriptors, desc)(mol, confId=conformer_id)
                except:
                    pass
                desc_val.extend(val)

        desc_val = to_numpy(desc_val)
        if self.replace_nan:
            desc_val = np.nan_to_num(desc_val)
        return desc_val



from typing import List
from typing import Union
from typing import Dict

import matplotlib.cm
import matplotlib.colors

import datamol as dm
import pandas as pd

from rdkit.Chem import rdMolChemicalFeatures


def colors_from_feature_factory(
    feature_factory: rdMolChemicalFeatures.MolChemicalFeatureFactory,
    cmap_name: str = "Set1",
    alpha: float = 1.0,
):
    """Get a list of colors for a given feature factory. For the same
    `feature_factory` the returned colors will be the same.

    Args:
        feature_factory: Feature factory to use.
        cmap_name: Matplotlib colormap name.
        alpha: Alpha value for the colors.

    Returns:
        colors: Dict of feature_name as keys and colors as values.
    """
    cmap_name = "Set1"

    cmap = matplotlib.cm.get_cmap(cmap_name)
    cmap_n = cmap.N  # type: ignore

    colors = {}
    for i, name in enumerate(feature_factory.GetFeatureFamilies()):
        color: List[float] = list(cmap(i % cmap_n))
        color[3] = alpha
        colors[name] = color

    return colors


def show_mols(mols: Union[dm.Mol, List[dm.Mol]]):
    """Generate a view of the molecules.

    Args:
        mols: A mol or a list of mols.

    Returns:
        nglview.widget.NGLWidget
    """

    import nglview as nv

    if isinstance(mols, dm.Mol):
        mols = [mols]

    view = nv.NGLWidget()
    for mol in mols:
        component = view.add_component(mol)
        component.clear()  # type: ignore
        component.add_ball_and_stick(multipleBond=True)  # type: ignore

    return view


def show_pharm_features(
    mols: Union[dm.Mol, List[dm.Mol]],
    features: pd.DataFrame,
    feature_factory: rdMolChemicalFeatures.MolChemicalFeatureFactory,
    alpha: float = 1.0,
    sphere_radius: float = 0.4,
    show_legend: bool = True,
):
    """Generate a view of the molecules with pharmacophoric features.

    Args:
        mols: A mol or a list of mols.
        features: Features data. Columns must contain at least
            "feature_name", "feature_id", and "feature_coords".
        feature_factory: Feature factory to display consistent colors.
        alpha: Alpha value for the colors (currently not working).
        sphere_radius: Radius of the spheres for the features.
        show_legend: Display the legend (the layout is bad but at least it
            shows the legend).

    Returns:
        nglview.widget.NGLWidget
    """

    import ipywidgets as ipy

    # Get mols view
    mol_view = show_mols(mols)

    # Get colors
    colors = colors_from_feature_factory(feature_factory, alpha=alpha)

    # Add features to the viz
    for _, row in features.iterrows():
        color = colors[row["feature_name"]]
        label = f"{row['feature_name']}_{row['feature_id']}"
        mol_view.shape.add_sphere(row["coords"], color, sphere_radius, label)  # type: ignore

    if not show_legend:
        return mol_view

    # Build legend widget
    colors_widget = _build_colors_widget(colors)

    main_layout = ipy.Layout(
        display="flex",
        flex_flow="column",
        align_content="center",
    )
    main_widget = ipy.HBox([mol_view, colors_widget], layout=main_layout)  # type: ignore

    return main_widget


def _build_colors_widget(colors: Dict[str, list]):
    import ipywidgets as ipy

    box_layout = ipy.Layout(
        display="flex",
        flex_flow="column",
        align_content="center",
        border="solid",
    )

    items = []
    for name, color in colors.items():
        item = ipy.Button(description=name)
        item.style.button_color = matplotlib.colors.to_hex(color)  # type: ignore
        items.append(item)

    box = ipy.Box(children=items, layout=box_layout)

    return box



from __future__ import annotations

from typing import List, Optional
from typing import Union
from typing import Optional

import os
import uuid
import torch
import tempfile
import numpy as np
import datamol as dm

from dataclasses import dataclass
from loguru import logger
from transformers import AutoTokenizer
from transformers import AutoModel
from transformers import AutoConfig
from transformers import MODEL_MAPPING
from transformers import PreTrainedModel
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from molfeat.trans.pretrained.base import PretrainedMolTransformer
from molfeat.store.loader import PretrainedStoreModel
from molfeat.store import ModelStore
from molfeat.store import ModelInfo
from molfeat.utils.converters import SmilesConverter
from molfeat.utils.pooler import get_default_hgf_pooler


@dataclass
class HFExperiment:
    model: PreTrainedModel
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
    notation: str = "smiles"

    @classmethod
    def save(cls, model: HFExperiment, path: str, clean_up: bool = False):
        """Save a hugging face model to a specific path

        Args:
            model: model to save
            path: path to the folder root where to save the model
            clean_up: whether to clean up the local path after saving
        """
        local_path = tempfile.mkdtemp()
        # we can save both the tokenizer and the model to the same path
        model.model.save_pretrained(local_path)
        model.tokenizer.save_pretrained(local_path)
        dm.fs.copy_dir(local_path, path, force=True, progress=True, leave_progress=False)
        logger.info(f"Model saved to {path}")
        # clean up now
        if clean_up:
            mapper = dm.fs.get_mapper(local_path)
            mapper.fs.delete(local_path, recursive=True)
        return path

    @classmethod
    def load(cls, path: str, model_class=None):
        """Load a model from the given path
        Args:
            path: Path to the model to load
            model_class: optional model class to provide if the model should be loaded with a specific class
        """
        if not dm.fs.is_local_path(path):
            local_path = tempfile.mkdtemp()
            dm.fs.copy_dir(path, local_path, force=True, progress=True, leave_progress=False)
        else:
            local_path = path

        if model_class is None:
            model_config = AutoConfig.from_pretrained(local_path)
            architectures = getattr(model_config, "architectures", [])
            if len(architectures) > 0:
                model_class = MODEL_MAPPING._load_attr_from_module(
                    model_config.model_type, architectures[0]
                )
            else:
                model_class = AutoModel
        model = model_class.from_pretrained(local_path)
        tokenizer = AutoTokenizer.from_pretrained(local_path)
        return cls(model, tokenizer)


class HFModel(PretrainedStoreModel):
    """Transformer model loading model loading"""

    def __init__(
        self,
        name: str,
        cache_path: Optional[os.PathLike] = None,
        store: Optional[ModelStore] = None,
    ):
        """Model loader initializer

        Args:
            name (str, optional): Name of the model for ada.
            cache_path (os.PathLike, optional): Local cache path for faster loading. This is the cache_path parameter for ADA loading !
        """

        super().__init__(name, cache_path=cache_path, store=store)
        self._model = None

    @classmethod
    def _ensure_local(cls, object_path: Union[str, os.PathLike]):
        """Make sure the input path is a local path otherwise download it

        Args:
            object_path: Path to the object

        """
        if dm.fs.is_local_path(object_path):
            return object_path
        local_path = tempfile.mkdtemp()
        if dm.fs.is_file(object_path):
            local_path = os.path.join(local_path, os.path.basename(object_path))
            dm.fs.copy_file(object_path, local_path)
        else:
            dm.fs.copy_dir(object_path, local_path)
        return local_path

    @classmethod
    def from_pretrained(
        cls,
        model: Union[str, PreTrainedModel],
        tokenizer: Union[str, PreTrainedTokenizer, PreTrainedTokenizerFast],
        model_class=None,
        model_name: Optional[str] = None,
    ):
        """Load model using huggingface pretrained model loader hook

        Args:
            model: Model to load. Can also be the name on the hub or the path to the model
            tokenizer: Tokenizer to load. Can also be the name on the hub or the path to the tokenizer
            model_class: optional model class to provide if the model should be loaded with a specific class
            model_name: optional model name to give to this model.
        """

        # load the model
        if isinstance(model, PreTrainedModel):
            model_obj = model
        else:
            if dm.fs.exists(model):
                model = cls._ensure_local(model)
            if model_class is None:
                model_config = AutoConfig.from_pretrained(model)
                architectures = getattr(model_config, "architectures", [])
                if len(architectures) > 0:
                    model_class = MODEL_MAPPING._load_attr_from_module(
                        model_config.model_type, architectures[0]
                    )
                else:
                    model_class = AutoModel
            model_obj = model_class.from_pretrained(model)

        if isinstance(tokenizer, (PreTrainedTokenizer, PreTrainedTokenizerFast)):
            tokenizer_obj = tokenizer
        else:
            if dm.fs.exists(tokenizer):
                tokenizer = cls._ensure_local(tokenizer)
            tokenizer_obj = AutoTokenizer.from_pretrained(tokenizer)
        name = model_name or f"hf_model_{uuid.uuid4().hex[:8]}"
        model = HFModel(name=name, store=ModelStore())
        model._model = HFExperiment(model=model_obj, tokenizer=tokenizer_obj)
        return model

    @classmethod
    def register_pretrained(
        cls,
        model: Union[str, PreTrainedModel],
        tokenizer: Union[str, PreTrainedTokenizer, PreTrainedTokenizerFast],
        model_card: ModelInfo,
        model_class=None,
    ):
        """Register a pretrained huggingface model to the model store
        Args:
            model: Model to load. Can also be the name on the hub or the path to the model
            tokenizer: Tokenizer to load. Can also be the name on the hub or the path to the tokenizer
            model_class: optional model class to provide if the model should be loaded with a specific class
            model_card: optional model card to provide for registering this model
        """
        model = cls.from_pretrained(model, tokenizer, model_class, model_name=model_card.name)
        model.store.register(model_card, model._model, save_fn=HFExperiment.save)
        return model

    def get_notation(self, default_notation: Optional[str] = None):
        """Get the notation of the model"""
        notation = default_notation
        try:
            modelcard = self.store.search(name=self.name)[0]
            notation = modelcard.inputs
        except Exception:
            pass
        return notation

    def load(self):
        """Load Transformer Pretrained featurizer model"""
        if self._model is not None:
            return self._model
        download_output_dir = self._artifact_load(
            name=self.name, download_path=self.cache_path, store=self.store
        )
        model_path = dm.fs.join(download_output_dir, self.store.MODEL_PATH_NAME)
        model = HFExperiment.load(model_path)
        return model


class PretrainedHFTransformer(PretrainedMolTransformer):
    """
    HuggingFace Transformer for feature extraction.

    !!! note
        For convenience and consistency, this featurizer only accepts as inputs
        smiles and molecules, then perform the internal conversion, based on the notation provided.
    """

    NEEDS_RANDOM_SEED = ["bert", "roberta"]

    def __init__(
        self,
        kind: Union[str, HFModel] = "ChemBERTa-77M-MLM",
        notation: Optional[str] = "none",
        pooling: str = "mean",
        concat_layers: Union[List[int], int] = -1,
        dtype=np.float32,
        device="cpu",
        max_length: int = 128,
        ignore_padding: bool = True,
        preload: bool = False,
        n_jobs: int = 0,
        random_seed: Optional[int] = None,
        **params,
    ):
        """
        HuggingFace Transformer for featurizer extraction
        The default behaviour of this feature extractor is to return the last hidden state of the encoder
        similar to what is performed by the pipeline 'feature-extraction' in hugging face.

        !!! warning
            For bert models, the default pooling layers is a neural network. Therefore, do not use the default
            Or provide a random seed for reproducibility (in this case pooling will act as random projection to the same manifold)

        !!! note
            The pooling module of this featurizer is accessible through the `_pooling_obj` attribute.

        Args:
            kind: name of the featurizer as available in the model store
            notation: optional line notation to use. Only use if it cannot be found from the model card.
            pooling: type of pooling to use. One of ['default', 'mean', 'max', 'sum']. The value "default" corresponds to the default litterature pooling for each model type.
                See `molfeat.utils.pooler.get_default_hf_pooler` for more details.
            concat_layers: Layer to concat to get the representation. By default the last hidden layer is returned.
            dtype: Data type to output
            device: Torch device on which to run the featurizer.
            max_length: Maximum length of the input sequence to consider. Please update this for large sequences
            ignore_padding: Whether to ignore padding in the representation (default: True) to avoid effect of batching
            preload: Whether to preload the model into memory or not
            n_jobs: number of jobs to use
            random_seed: random seed to use for reproducibility whenever a DNN pooler is used (e.g bert/roberta)
        """

        super().__init__(
            dtype=dtype,
            device=device,
            n_jobs=n_jobs,
            **params,
        )
        if concat_layers is None:
            concat_layers = -1
        if not isinstance(concat_layers, list):
            concat_layers = [concat_layers]
        self.concat_layers = concat_layers
        self.max_length = max_length
        self.ignore_padding = ignore_padding
        self._require_mols = False
        self.random_seed = random_seed
        self.preload = preload
        self.pooling = pooling
        self._pooling_obj = None
        if isinstance(kind, HFModel):
            self.kind = kind.name
            self.featurizer = kind
        else:
            self.kind = kind
            self.featurizer = HFModel(name=self.kind)
        self.notation = self.featurizer.get_notation(notation) or "none"
        self.converter = SmilesConverter(self.notation)
        if self.preload:
            self._preload()

    def _update_params(self):
        """Update the parameters of this model"""
        # pylint: disable=no-member
        super()._update_params()

        hf_model = HFModel(
            name=self.kind,
        )
        self.featurizer = hf_model.load()
        config = self.featurizer.model.config.to_dict()
        self._pooling_obj = self._pooling_obj = (
            get_default_hgf_pooler(self.pooling, config, random_seed=self.random_seed)
            if self._pooling_obj is None
            else self._pooling_obj
        )

    def _preload(self):
        """Perform preloading of the model from the store"""
        super()._preload()
        self.featurizer.max_length = self.max_length

        # we can be confident that the model has been loaded here
        if self._pooling_obj is not None and self.preload:
            return
        config = self.featurizer.model.config.to_dict()
        cur_tokenizer = self.featurizer.tokenizer
        for special_token_id_name in [
            "pad_token_id",
            "bos_token_id",
            "eos_token_id",
            "unk_token_id",
            "sep_token_id",
            "mask_token_id",
        ]:
            token_id = getattr(cur_tokenizer, special_token_id_name)
            if token_id is not None:
                config[special_token_id_name] = token_id

        self._pooling_obj = (
            get_default_hgf_pooler(self.pooling, config, random_seed=self.random_seed)
            if self._pooling_obj is None
            else self._pooling_obj
        )
        # pooling layer is still none, that means we could not fetch it properly
        if self._pooling_obj is None:
            logger.warning(
                "Cannot confidently find the pooling layer and therefore will not apply pooling"
            )

    def _convert(self, inputs: list, **kwargs):
        """Convert the list of molecules to the right format for embedding

        Args:
            inputs: inputs to preprocess

        Returns:
            processed: pre-processed input list
        """
        self._preload()

        if isinstance(inputs, (str, dm.Mol)):
            inputs = [inputs]

        def _to_smiles(x):
            return dm.to_smiles(x) if not isinstance(x, str) else x

        parallel_kwargs = getattr(self, "parallel_kwargs", {})

        if len(inputs) > 1:
            smiles = dm.utils.parallelized(
                _to_smiles,
                inputs,
                n_jobs=self.n_jobs,
                **parallel_kwargs,
            )
            inputs = dm.utils.parallelized(
                self.converter.encode,
                smiles,
                n_jobs=self.n_jobs,
                **parallel_kwargs,
            )
        else:
            inputs = self.converter.encode(_to_smiles(inputs[0]))
        # this check is necessary for some tokenizers
        if isinstance(inputs, str):
            inputs = [inputs]
        encoded = self.featurizer.tokenizer(
            list(inputs),
            truncation=True,
            padding=True,
            max_length=self.max_length,
            return_tensors="pt",
        )
        return encoded

    def _embed(self, inputs, **kwargs):
        """
        Perform embedding of inputs using the pretrained model

        Args:
            inputs: smiles or seqs
            kwargs: any additional parameters
        """
        self._preload()

        attention_mask = inputs.get("attention_mask", None)
        if attention_mask is not None and self.ignore_padding:
            attention_mask = attention_mask.unsqueeze(-1)  # B, S, 1
        else:
            attention_mask = None
        with torch.no_grad():
            out_dict = self.featurizer.model(output_hidden_states=True, **inputs)
            hidden_state = out_dict["hidden_states"]
            emb_layers = []
            for layer in self.concat_layers:
                emb = hidden_state[layer].detach().cpu()  # B, S, D
                emb = self._pooling_obj(
                    emb,
                    inputs["input_ids"],
                    mask=attention_mask,
                    ignore_padding=self.ignore_padding,
                )
                emb_layers.append(emb)
            emb = torch.cat(emb_layers, dim=1)
        return emb.numpy()

    def set_max_length(self, max_length: int):
        """Set the maximum length for this featurizer"""
        self.max_length = max_length
        self._preload()



from typing import Union
from typing import Iterator
from typing import Iterable
from typing import Callable
from typing import Optional
from typing import Dict
from typing import List
from typing import Any

import numpy as np
import pandas as pd
import datamol as dm

from rdkit.Chem import rdchem
from sklearn.base import BaseEstimator
from molfeat.trans.fp import FPVecTransformer
from molfeat.utils import datatype


class FeatConcat(list, BaseEstimator):
    r"""
    Concatenation container for `FPVecTransformer`. This class allows
    merging multiple fingerprints into a single one.
    It gives the ability to call the following methods
        - `fit`
        - `transform`
        - `fit_transform`
    on a list of transformers and concatenate the results.

    !!! note
        The featurization length of this featurizer is accessible via the `length` property.
        `len()` will return the number of base featurizer.
    """

    _STR_DELIMITER = "||"

    def __init__(
        self,
        iterable: Optional[Union[Iterable, str]] = None,
        dtype: Optional[Callable] = None,
        params: Optional[Dict[str, Any]] = None,
        collate_fn: Optional[Callable] = None,
    ):
        """Featurizer concatenator

        Args:
            iterable: List of featurizer to concatenate.
            dtype: Datatype of the computed fingerprint
            params: Optional dictionary of parameters for the featurizers when there is a need for initializing them
            collate_fn: optional function to provide for custom collating.
                By default the collate function will be None, which will use the torch default
        """
        super().__init__()
        self.params = params or {}
        if isinstance(iterable, str):
            iterable = [x.strip() for x in iterable.split(self._STR_DELIMITER)]
        if iterable is not None:
            for item in iterable:
                if isinstance(item, str):
                    item = FPVecTransformer(kind=item, **self.params.get(item, {}))
                self.append(item)
        self.dtype = dtype
        self._length = None
        self.collate_fn = collate_fn

    def append(self, item):
        r"""Override the ``append`` to accept only ``FPVecTransformer``"""
        self._check_supported(item)
        super().append(item)

    def insert(self, index, item):
        r"""Override the ``insert`` to accept only ``BaseFeaturizer``"""
        self._check_suported(item)
        super().insert(index, item)

    def __add__(self, item):
        """Override the `__add__` method"""
        self._check_supported(item)
        super().__add__(item)

    def __setitem__(self, index, value):
        """Override the `__setitem__`  method"""
        self._check_supported(value)
        super().__setitem__(index, value)

    @property
    def length(self):
        """
        Length property for Feat concatenation.  This is the sum of the length of each transformer.
        Note that __len__ returns the number of base featurizer here instead.
        """
        if self._length is None:
            full_length = 0
            for feat in self:
                if isinstance(feat, FeatConcat):
                    full_length += feat.length
                else:
                    full_length += len(feat)
            self._length = full_length
        return self._length

    def _check_supported(self, item):
        r"""Check if the item is the right type"""
        if not isinstance(item, FPVecTransformer):
            raise ValueError("FPVecTransformer allowed only, provided {}".format(type(item)))

    def get_collate_fn(self, *args, **kwargs):
        """
        Get collate function of this featurizer. The FeatConcat featurizer use the default
        collate function which does not do anything.

        Returns:
            fn: Collate function for pytorch or None
        """
        return getattr(self, "collate_fn", None)

    def iter_index(self, indexes: Union[int, Iterator[int]]):
        r"""
        Allow the `FeatConcat` to be indexed using a list, or any other iterable.

        Args:
            indexes: The indexes to index the ``FeatConcat``.

        Returns
            indexed_fpconcat: A new FeatConcat object with the indexed element
        """
        if not isinstance(indexes, (list, tuple)):
            try:
                indexes = list(indexes)
            except TypeError:
                indexes = [indexes]
        return FeatConcat([self[ii] for ii in indexes])

    @property
    def columns(self):
        """Get the list of columns for the concatenated molecules

        Returns:
            columns (list): Name of the columns of the descriptor
        """
        tmp_mol = dm.to_mol("CC(C)O")
        columns = []
        for fp in self:
            fp_columns = getattr(fp, "columns", None)
            fp_name = str(fp)
            if fp_columns is None:
                fp_out, _ = fp([tmp_mol])
                fp_out = np.asarray(fp_out)
                fp_columns = [f"{fp_name}:{ind}" for ind in range(fp_out.shape[-1])]
            columns.extend(fp_columns)
        return columns

    def transform(self, mols: List[Union[rdchem.Mol, str]], **kwargs):
        r"""
        Calls the ``FPVecTransformer.transform`` for each transformer in
        the current list, and concatenates the resulting fingerprints.

        Args:
            mols: List of SMILES or molecules
            kwargs: named parameters for transform (see below)

        Returns:
            fps: Computed fingerprints of size NxD, where D is the
                sum of the length of each transformer and N is the number of input
                molecules that have been successfully featurized.
        """

        fps = []
        for _, fp_trans in enumerate(self):
            out = fp_trans.transform(mols, enforce_dtype=False, **kwargs)
            out = datatype.cast(out, dtype="pandas")
            fps.append(out)
        fps = pd.concat(fps, axis=1)
        fps.columns = self.columns
        return fps.values

    def __call__(
        self,
        mols: List[Union[rdchem.Mol, str]],
        enforce_dtype: bool = False,
        ignore_errors: bool = False,
        **kwargs,
    ):
        r"""
        Calls each of the internal transformer and concatenate results only on valid indices.

        Args:
            mols:  Mol or SMILES of the molecules to be transformed
            enforce_dtype: whether to enforce the instance dtype in the generated fingerprint
            ignore_errors: Whether to ignore errors during featurization or raise an error.
            kwargs: Named parameters for the transform method

        Returns:

            fp: array
                computed fingerprints of size NxD, where D is the
                sum of the length of each transformer and N is the number of input
                molecules that have been successfully featurized.
            idx: array
                Indices of successful featurization given the original molecule input.
        """

        fps = []
        valid_idx = np.zeros(len(mols))
        for _, transf in enumerate(self):
            fp = transf.transform(mols, ignore_errors=ignore_errors, **kwargs)
            fp, idx = transf._filter_none(fp)
            fps.append(fp)
            valid_idx[idx] += 1  # increase counter of valids
        valid_idx = np.nonzero(valid_idx == len(self))[0]
        fps = np.concatenate(fps, axis=1)
        if self.dtype is not None and enforce_dtype:
            fps = datatype.cast(fps, dtype=self.dtype, columns=self.columns)
        if not ignore_errors:
            return fps
        return fps, list(valid_idx)

    def fit_transform(
        self,
        mols: List[Union[str, rdchem.Mol]],
        y: Optional[Iterable] = None,
        fit_kwargs: Dict = None,
        trans_kwargs: Dict = None,
    ):
        r"""
        Calls the ``self.fit`` followed by the ``fit.transform`` for each transfomer in
        the current list, and concatenates the resulting fingerprints.

        Args:
            mols: List of SMILES or molecules
            y: target for the fitting. Usually ignored for FPVecTransformer
            fit_kwargs:  named parameters for fit
            fit_kwargs:named parameters for transform

        Returns:

            fp: computed fingerprints of size NxD, where D is the
                sum of the length of each transformer and N is the number of input
                molecules that have been successfully featurized.
        """
        fit_kwargs = {} if fit_kwargs is None else fit_kwargs
        trans_kwargs = {} if trans_kwargs is None else trans_kwargs
        self.fit(mols, y=y, **fit_kwargs)
        return self.transform(mols, **trans_kwargs)

    def fit(self, X: List[Union[rdchem.Mol, str]], y=None, **kwargs):
        r"""
        Calls the ``FPVecTransformer.fit`` for each transformer in the current list.

        Args:
            X: input list of molecules
            y (list, optional): Optional list of molecular properties. Defaults to None.

        Returns:
            self: FeatConcat instance after fitting
        """

        for _, fp_trans in enumerate(self):
            fp_trans.fit(X, y=y, **kwargs)
        return self
